"""Implementation of a Gated ReLU layer in PyTorch."""

import torch
import torch.nn.functional as F


class GatedReLULayer(torch.nn.Module):

    """A gated ReLU layer with activation.

        Z = X W * max(X U, 0),

    where 'U' is a fixed, pre-specified set of gate vectors.
    """

    def __init__(self, U: torch.Tensor):
        """
        :param U: the gates vectors for the gated ReLU activation.
        """
        super().__init__()
        self.U = U

        # infer input shape and layer width from gate vectors
        self.in_features, self.out_features = self.U.shape

        # linear layer
        self.linear = torch.nn.Linear(self.in_features, self.out_features, bias=False)

    def forward(self, x: torch.Tensor, batch_size=None):
        """
        :param x: the input data.
        :returns: the activations.
        """
        # return linear combinations gated by the U vectors.
        return torch.multiply(torch.sign(F.relu(x @ self.U)), self.linear(x))



"""Non-convex models implemented in PyTorch."""

from .sequential_wrapper import SequentialWrapper
from .layer_wrapper import LayerWrapper
from .gated_relu_layer import GatedReLULayer

__all__ = [
    "SequentialWrapper",
    "LayerWrapper",
    "GatedReLULayer",
]



"""Utilities for initializing optimization methods."""

from typing import Union, Tuple, Optional, List
import math

from scipy.sparse.linalg import eigsh  # type: ignore

import lab

from scnn.private.models.model import Model


def init_batch_size(
    train_set: Tuple[lab.Tensor, lab.Tensor], batch_size: int = 128
) -> int:
    """
    :param train_set: an (X,y) tuple containing the training set.
    :param batch_size: the batch size to use. If less than 1, it is assumed that the
        provided floating point number is the fraction of the training set which the
        the mini-batch should compose.
    :returns: the desired batch size.
    """
    # only floating point or integer numbers are permitted.
    assert isinstance(batch_size, float) or isinstance(batch_size, int)

    if batch_size == -1:  # full batch
        return train_set[0].shape[0]
    elif batch_size < 1.0:
        return max(int(math.floor(train_set[0].shape[0] * batch_size)), 1)
    else:
        return batch_size


def init_max_epochs(
    train_set: Tuple[lab.Tensor, lab.Tensor],
    max_epochs: Optional[int] = None,
    max_iters: Optional[int] = None,
    batch_size: Optional[int] = None,
) -> int:
    """Determine the maximum number of epochs to run an optimization algorithm
    given either either the number of epochs directly or the number of
    iterations desired.

    :param train_set: an (X,y) tuple containing the training set.
    :param max_epochs: the number of epochs to run. This will be return directly if it is non 'None'.
    :param max_iters: the maximum number of (inner) iterations to run the optimization algorithm.
        This can be useful to provide if the goal is to run a fixed number of iterations regardless
        of the dataset size.
    :param batch_size: the batch size that will be used by the optimizer.
        This must be provided if 'max_iters' is not None.
    """
    # only one of max_epochs or max_iters may be provided.
    assert max_epochs is None or max_iters is None

    if max_epochs is not None:
        return max_epochs
    elif batch_size is None:
        return max_iters
    else:
        return int(math.ceil(max_iters * batch_size / train_set[0].shape[0]))



"""Optimize neural networks using convex reformulations.

TODO:
    - extract types into a new `types.py` module.

    - use_bias: update mapping for convex/non-convex models to support bias
        terms without data augmentation.

"""
import math
import os
import pickle as pkl
from copy import deepcopy
from typing import List, Literal, Optional, Tuple, Union

import numpy as np

from scnn.activations import sample_gate_vectors
from scnn.metrics import Metrics
from scnn.models import ConvexGatedReLU, ConvexReLU, Model
from scnn.private.interface import (
    build_internal_model,
    build_internal_regularizer,
    build_metrics_tuple,
    build_optimizer,
    build_public_model,
    get_logger,
    normalized_into_input_space,
    process_data,
    set_device,
    update_public_metrics,
    update_public_model,
)
from scnn.private.models.solution_mappings import get_nc_formulation
from scnn.regularizers import Regularizer
from scnn.solvers import (
    AL,
    RFISTA,
    ApproximateConeDecomposition,
    Optimizer,
)

# Types

Formulation = Literal["gated_relu", "relu"]
Device = Literal["cpu", "cuda"]
Dtype = Literal["float32", "float64"]


def optimize(
    formulation: Formulation,
    max_neurons: int,
    X_train: np.ndarray,
    y_train: np.ndarray,
    X_test: Optional[np.ndarray] = None,
    y_test: Optional[np.ndarray] = None,
    regularizer: Optional[Regularizer] = None,
    bias: bool = False,
    return_convex: bool = False,
    unitize_data: bool = True,
    verbose: bool = False,
    log_file: str = None,
    device: Device = "cpu",
    dtype: Dtype = "float32",
    seed: int = 778,
) -> Tuple[Model, Metrics]:
    """Convenience function for training neural networks by convex
    reformulation.

    Args:
        formulation: the convex reformulation to solve. Must be one of

            - `"gated_relu"`: train a network with Gated ReLU activations.

            - `"relu"`: train a network with ReLU activations.

        max_neurons: the maximum number of neurons in the convex reformulation.
        X_train: an :math:`n \\times d` matrix of training examples.
        y_train: an :math:`n \\times c` or vector matrix of training targets.
        X_test: an :math:`m \\times d` matrix of test examples.
        y_test: an :math:`n \\times c` or vector matrix of test targets.
        regularizer: an optional regularizer for the convex reformulation.
            Defaults to no regularization.
        bias: whether or not to use a bias in the model.
        return_convex: whether or not to return the convex reformulation
            instead of the final non-convex model.
        unitize_data: whether or not to unitize the column norms of the
            training set. This can improve conditioning during optimization.
        verbose: whether or not the solver should print verbosely during
            optimization.
        log_file: a path to an optional log file.
        device: the device on which to run. Must be one of `cpu` (run on CPU)
            or `cuda` (run on cuda-enabled GPUs if available).
        dtype: the floating-point type to use. `"float32"` is faster than
            `"float64"` but can lead to excessive numerical errors on badly
            conditioned datasets.
        seed: an integer seed for reproducibility.

    Returns:
        (Model, Metrics): the optimized model and metrics collected during
        optimization.
    """

    model: Model
    solver: Optimizer

    d = len(X_train[0])
    # check number of outputs
    try:
        c = len(y_train[0])
    except TypeError:
        c = 1

    # Instantiate convex model and other options.
    if formulation == "gated_relu":
        G = sample_gate_vectors(seed, d, max_neurons)
        model = ConvexGatedReLU(G, c=c, bias=bias)
        solver = RFISTA(model)
    elif formulation == "relu":
        # ReLU models can have up to 2 * G neurons.
        G = sample_gate_vectors(seed, d, math.floor(max_neurons / 2))
        model = ConvexReLU(G, c=c, bias=bias)
        solver = AL(model)
    else:
        raise ValueError(f"Convex formulation {formulation} not recognized!")

    metrics = Metrics()

    return optimize_model(
        model,
        solver,
        metrics,
        X_train,
        y_train,
        X_test,
        y_test,
        regularizer,
        return_convex,
        unitize_data,
        verbose,
        log_file,
        device,
        dtype,
        seed,
    )


def optimize_model(
    model: Model,
    solver: Optimizer,
    metrics: Metrics,
    X_train: np.ndarray,
    y_train: np.ndarray,
    X_test: Optional[np.ndarray] = None,
    y_test: Optional[np.ndarray] = None,
    regularizer: Optional[Regularizer] = None,
    return_convex: bool = False,
    unitize_data: bool = True,
    verbose: bool = False,
    log_file: str = None,
    device: Device = "cpu",
    dtype: Dtype = "float32",
    seed: int = 778,
) -> Tuple[Model, Metrics]:
    """Train a neural network by convex reformulation.

    Args:
        model: a convex reformulation of a neural network model.
        solver: the optimizer to use when solving the reformulation.
        metrics: a object specifying which metrics to collect during
            optimization.
        X_train: an :math:`n \\times d` matrix of training examples.
        y_train: an :math:`n \\times c` or vector matrix of training targets.
        X_test: an :math:`m \\times d` matrix of test examples.
        y_test: an :math:`n \\times c` or vector matrix of test targets.
        regularizer: an optional regularizer for the convex reformulation.
        return_convex: whether or not to return the convex reformulation
            instead of the final non-convex model.
        unitize_data: whether or not to unitize the column norms of the
            training set. This can improve conditioning during optimization.
        verbose: whether or not the solver should print verbosely during
            optimization.
        log_file: a path to an optional log file.
        device: the device on which to run. Must be one of `cpu` (run on CPU)
            or `cuda` (run on cuda-enabled GPUs if available).
        dtype: the floating-point type to use. `"float32"` is faster than
            `"float64"` but can lead to excessive numerical errors on badly
            conditioned datasets.
        seed: an integer seed for reproducibility.

    Returns:
        The optimized model and metrics collected during optimization.
    """
    logger = get_logger("scnn", verbose, False, log_file)

    # set backend settings.
    if solver.cpu_only() and device != "cpu":
        logger.warning(
            f"Solver {solver} only supports CPU. User supplied device {device}\
            has been overridden."
        )
        device = "cpu"

    set_device(device, dtype, seed)

    if metrics.has_test_metrics() and (X_test is None or y_test is None):
        logger.warning(
            "Metrics specifies test metrics, but no test set was provided. \
            Test metrics will be collected on the training set. \n"
        )

    # Note: this unitizes columns of data matrix.
    (X_train, y_train), (X_test, y_test), column_norms = process_data(
        X_train,
        y_train,
        X_test,
        y_test,
        unitize_data,
        model.bias,
    )

    internal_model = build_internal_model(model, regularizer, X_train)
    opt_procedure = build_optimizer(solver, regularizer, metrics)
    metrics_tuple = build_metrics_tuple(metrics)

    initializer = lambda model: model

    exit_status, internal_model, internal_metrics = opt_procedure(
        logger,
        internal_model,
        initializer,
        (X_train, y_train),
        (X_test, y_test),
        metrics_tuple,
    )

    metrics = update_public_metrics(metrics, internal_metrics)

    # convert internal metrics

    # transform model back to original data space if necessary.
    if unitize_data:
        internal_model.weights = normalized_into_input_space(
            internal_model.weights, column_norms
        )

    if return_convex:
        return update_public_model(model, internal_model), metrics

    # convert into internal non-convex model
    nc_internal_model = get_nc_formulation(internal_model, remove_sparse=True)

    # create non-convex model
    return build_public_model(nc_internal_model, model.bias), metrics


def optimize_path(
    model: Model,
    solver: Optimizer,
    path: List[Regularizer],
    metrics: Metrics,
    X_train: np.ndarray,
    y_train: np.ndarray,
    X_test: Optional[np.ndarray] = None,
    y_test: Optional[np.ndarray] = None,
    warm_start: bool = True,
    save_path: Optional[str] = None,
    return_convex: bool = False,
    unitize_data: bool = True,
    verbose: bool = False,
    log_file: str = None,
    device: Device = "cpu",
    dtype: Dtype = "float32",
    seed: int = 778,
) -> Tuple[List[Union[Model, str]], List[Metrics]]:
    """Train a neural network by convex reformulation.

    Args:
        model: a convex reformulation of a neural network model.
        solver: the optimizer to use when solving the reformulation.
        path: a list of regularizer objects specifying the regularization path
            to explore.
        metrics: a object specifying which metrics to collect during
            optimization.
        X_train: an :math:`n \\times d` matrix of training examples.
        y_train: an :math:`n \\times c` or vector matrix of training targets.
        X_test: an :math:`m \\times d` matrix of test examples.
        y_test: an :math:`n \\times c` or vector matrix of test targets.
        warm_start: whether or not to warm-start each optimization problem in
            the path using the previous solution.
        save_path: string specifying a directory where models in the
            regularization path should be saved. All models will be retained
            in memory and returned if `save_path = None`.
        return_convex: whether or not to return the convex reformulation
            instead of the final non-convex model.
        unitize_data: whether or not to unitize the column norms of the
            training set. This can improve conditioning during optimization.
        verbose: whether or not the solver should print verbosely during
            optimization.
        log_file: a path to an optional log file.
        device: the device on which to run. Must be one of `cpu` (run on CPU)
            or `cuda` (run on cuda-enabled GPUs if available).
        dtype: the floating-point type to use. `"float32"` is faster than
            `"float64"` but can lead to excessive numerical errors on badly
            conditioned datasets.
        seed: an integer seed for reproducibility.
    """
    # set backend settings.
    logger = get_logger("scnn", verbose, False, log_file)

    # set backend settings.
    if solver.cpu_only() and device != "cpu":
        logger.warning(
            f"Solver {solver} only supports CPU. User supplied device {device}\
              has been overridden."
        )
        device = "cpu"

    set_device(device, dtype, seed)

    if metrics.has_test_metrics() and (X_test is None or y_test is None):
        logger.warning(
            "Metrics specifies test metrics, but no test set was provided.\
            Test metrics will be collected on the training set. \n"
        )

    # Note: this unitizes columns of data matrix.
    (X_train, y_train), (X_test, y_test), column_norms = process_data(
        X_train,
        y_train,
        X_test,
        y_test,
        unitize_data,
    )

    internal_model = build_internal_model(model, path[0], X_train)
    initializer = lambda model: model

    metrics_list: List[Metrics] = []
    model_list: List[Union[Model, str]] = []

    for regularizer in path:
        # update internal regularizer
        internal_model.regularizer = build_internal_regularizer(regularizer)
        opt_procedure = build_optimizer(solver, regularizer, metrics)

        metrics_tuple = build_metrics_tuple(metrics)

        exit_status, internal_model, internal_metrics = opt_procedure(
            logger,
            internal_model,
            initializer,
            (X_train, y_train),
            (X_test, y_test),
            metrics_tuple,
        )

        # regularizer to_string to generate path
        metrics = update_public_metrics(metrics, internal_metrics)
        cur_weights = internal_model.weights
        # transform model back to original data space if necessary.
        if unitize_data:
            internal_model.weights = normalized_into_input_space(
                internal_model.weights, column_norms
            )

        if return_convex:
            model_to_save = update_public_model(model, internal_model)
        else:
            nc_internal_model = get_nc_formulation(internal_model, remove_sparse=True)
            model_to_save = build_public_model(nc_internal_model, model.bias)

        if save_path is not None:

            reg_path = os.path.join(save_path, str(regularizer))
            os.makedirs(save_path, exist_ok=True)
            with open(reg_path, "wb") as f:
                pkl.dump((model_to_save, metrics), f)
            model_list.append(reg_path)

        else:
            model_list.append(deepcopy(model_to_save))

        metrics_list.append(deepcopy(metrics))
        internal_model.weights = cur_weights

    return model_list, metrics_list



"""Implementation of linear regression with the least-squares objective."""
from typing import Optional

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
import scnn.private.loss_functions as loss_fns


class LinearRegression(Model):

    """Linear regression with least-squares objective.

    Attributes:
        d: the input dimension.
        c: the output dimension.
        weights: an (c, d) matrix of weights for the linear model.
    """

    def __init__(
        self, d: int, c: int = 1, regularizer: Optional[Regularizer] = None
    ):
        """
        Args:
            d: the input dimension.
            c: the output dimension.
            regularizer: a penalty function controlling the flexibility of the
                model.
        """
        super().__init__(regularizer)
        self.d = d
        self.c = c
        self.weights = lab.zeros((c, d))

    def _forward(self, X: lab.Tensor, w: lab.Tensor, **kwargs) -> lab.Tensor:
        """Compute forward pass.

        Args:
            X: (n,d) array containing the data examples.
            w: parameter at which to compute the forward pass.

        Returns:
            The forward pass (i.e. predictions) at `w`.
        """
        return X @ w.T

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute least-squares objective given a dataset.

        Args:
            X: (n,d) array containing the data examples.
            y: (n,c) array containing the data targets.
            w: specific parameter at which to compute the forward pass.
            scaling: scaling parameter for the objective. Defaults to `n * c`.

        Returns:
            The objective at `w`.
        """
        return loss_fns.squared_error(self._forward(X, w), y) / (
            2 * self._scaling(y, scaling)
        )

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute gradient of the least-squares objective.

        Args:
            X: (n,d) array containing the data examples.
            y: (n,c) array containing the data targets.
            w: parameter at which to compute the forward pass.
            scaling: scaling parameter for the objective. Defaults to `n * c`.

        Returns:
            The gradient at `w`.
        """

        res = self._forward(X, w) - y
        grad = lab.matmul(res.T, X) / self._scaling(y, scaling)
        return grad



"""Non-convex and convex formulations of two-layer neural networks.

Overview:
    This module provides implementations of non-convex and convex formulations
    for two-layer ReLU and Gated ReLU networks. The difference between ReLU
    and Gated ReLU networks is the activation function; Gated ReLU networks
    use fixed "gate" vectors when computing the activation pattern while
    standard ReLU networks use the model parameters. Concretely, the prediction
    function for a two ReLU network is

    .. math:: h(X) = \\sum_{i=1}^p (X W_{1i}^{\\top})_+ \\cdot W_{2i}^{\\top},

    where :math:`W_{1} \\in \\mathbb{R}^{p \\times d}` are the parameters of
    the first layer, and :math:`W_{2} \\in \\mathbb{R}^{c \\times p}` are the
    parameters of the second layer. In contrast, Gated ReLU networks predict as

    .. math:: h(X) = \\sum_{i=1}^p \\text{diag}(X g_i > 0) X W_{1i}^{\\top}
        \\cdot W_{2i}^{\\top},

    where the :math:`g_i` vectors are fixed (ie. not learned) gates.

    The convex reformulations of the ReLU and Gated ReLU models are obtained
    by enumerating the possible activation patterns
    :math:`D_i = \\text{diag}(1(X g_i > 0))`.  For a Gated ReLU model, the
    activations are exactly specified by the set of gate vectors, while for
    ReLU models the space of activation is much larger.
    Using a (possibly subsampled) set of activations :math:`\\mathcal{D}`,
    the prediction function for the convex reformulation of a two-layer ReLU
    network can be written as

    .. math:: g(X) = \\sum_{D_i \\in \\mathcal{D}}^m D_i X (v_{i} - w_{i}),

    where :math:`v_i, w_i \\in \\mathbb{R}^{m \\times d}` are the model
    parameters. For Gated ReLU models, the convex reformulation is

    .. math:: g(X) = \\sum_{i=1}^m \\text{diag}(X g_i > 0) X U_{i},

    where :math:`U \\in \\mathbb{R}^{m \\times d}` are the model parameters
    and :math:`g_i` are the gate vectors from the non-convex model. For both
    convex reformulations, a one-vs-all strategy is used for the convex
    reformulation when the output dimension satisfies :math:`c > 1`.
"""

from typing import List, Optional

import numpy as np


class Model:
    """Base class for convex and non-convex models.

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons.
        bias: whether or not the model uses a bias term.
        parameters: a list of NumPy arrays comprising the model parameters.
    """

    d: int
    p: int
    c: int
    bias: bool
    parameters: List[np.ndarray]


class GatedModel(Model):
    """Abstract class for models with fixed gate vectors.

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons. This is is always `1` for a linear model.
        bias: whether or not the model uses a bias term.
        G: the gate vectors for the Gated ReLU activation stored as a
            (d x p) matrix.
        G_bias: an optional vector of biases for the gates.
    """

    def __init__(
        self,
        G: np.ndarray,
        c: int,
        bias: bool = False,
        G_bias: Optional[np.ndarray] = None,
    ):
        """Construct a new convex Gated ReLU model.

        Args:
            G: a (d x p) matrix of get vectors, where p is the
                number neurons.
            c: the output dimension.
            bias: whether or not to include a bias term.
            G_bias: a vector of bias parameters for the gates.
                Note that `bias` must be True for this to be supported.
        """
        self.G = G
        self.d, self.p = G.shape
        self.c = c
        self.bias = bias

        if bias is None:
            assert G_bias is None
        self.G_bias = G_bias

        if self.G_bias is None:
            self.G_bias = np.zeros(self.p)

    def compute_activations(self, X: np.ndarray) -> np.ndarray:
        """Compute activations for models with fixed gate vectors.

        Args:
            X: (n x d) matrix of input examples.

        Returns:
            D: (n x p) matrix of activation patterns.
        """
        D = np.maximum(X @ self.G + self.G_bias, 0)
        D[D > 0] = 1

        return D


class LinearModel(Model):
    """Basic linear model.

    This model has the prediction function :math:`g(X) = X W^\\top`, where
    :math:`W \\in \\mathbb{R}^{c \\times d}` is a matrix of weights.


    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons. This is is always `1` for a linear model.
        bias: whether or not the model uses a bias term.
        parameters: a list of NumPy arrays comprising the model parameters.
    """

    def __init__(self, d: int, c: int, bias: bool = False):
        """
        Args:
            d: the input dimension.
            c: the output dimension.
            bias: whether or not to include a bias term.
        """
        self.d = d
        self.c = c
        self.p = 1
        self.bias = bias

        if self.bias:
            self.parameters = [np.zeros((c, d)), np.zeros((1))]
        else:
            self.parameters = [np.zeros((c, d))]

    def get_parameters(self) -> List[np.ndarray]:
        """Get the model parameters."""
        return self.parameters

    def set_parameters(self, parameters: List[np.ndarray]):
        """Set the model parameters.

        This method safety checks the dimensionality of the new parameters.

        Args:
            parameters: the new model parameters.
        """
        assert parameters[0].shape == (self.c, self.d)

        if self.bias:
            assert parameters[1].shape == (1)

        self.parameters = parameters

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Compute the model predictions for a given dataset.

        Args:
            X: an (n  d) array containing the data examples on
                which to predict.

        Returns:
            - g(X): the model predictions for X.
        """
        y_hat = X @ self.parameters[0].T

        if self.bias:
            y_hat = y_hat + self.parameters[1]

        return y_hat


class ConvexGatedReLU(GatedModel):
    """Convex reformulation of a Gated ReLU Network with two-layers.

    This model has the prediction function

    .. math::

        g(X) = \\sum_{i=1}^m \\text{diag}(X g_i > 0) X U_{1i}.

    A one-vs-all strategy is used to extend the model to multi-dimensional
    targets.

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons.
        bias: whether or not the model uses a bias term.
        G: the gate vectors for the Gated ReLU activation stored as a
            (d x p) matrix.
        G_bias: an optional vector of biases for the gates.
        parameters: the parameters of the model stored as a list of tensors.
    """

    def __init__(
        self,
        G: np.ndarray,
        c: int = 1,
        bias: bool = False,
        G_bias: Optional[np.ndarray] = None,
    ) -> None:
        """Construct a new convex Gated ReLU model.

        Args:
            G: a (d x p) matrix of get vectors, where p is the
                number neurons.
            c: the output dimension.
            bias: whether or not to include a bias term.
            G_bias: a vector of bias parameters for the gates.
                Note that `bias` must be True for this to be supported.
        """

        super().__init__(G, c, bias, G_bias)

        # one linear model per gate vector
        if self.bias:
            self.parameters = [
                np.zeros((c, self.p, self.d)),
                np.zeros((c, self.p)),
            ]
        else:
            self.parameters = [np.zeros((c, self.p, self.d))]

    def get_parameters(self) -> List[np.ndarray]:
        """Get the model parameters."""
        return self.parameters

    def set_parameters(self, parameters: List[np.ndarray]):
        """Set the model parameters.

        This method safety checks the dimensionality of the new parameters.

        Args:
            parameters: the new model parameters.
        """
        assert parameters[0].shape == (self.c, self.p, self.d)

        if self.bias:
            assert parameters[1].shape == (self.c, self.p)

        self.parameters = parameters

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Compute the model predictions for a given dataset.

        Args:
            X: an (n  d) array containing the data examples on
                which to predict.

        Returns:
            - g(X): the model predictions for X.
        """
        D = super().compute_activations(X)

        if self.bias:
            Z = (
                np.einsum("ij, lkj->lik", X, self.parameters[0])
                + self.parameters[1]
            )

            # TODO: need to check this computation.
            return np.einsum("lik, ik->il", Z, D)
        else:
            return np.einsum("ij, lkj, ik->il", X, self.parameters[0], D)


class NonConvexGatedReLU(GatedModel):
    """Convex reformulation of a Gated ReLU Network with two-layers.

    This model has the prediction function

    .. math:: h(X) = \\sum_{i=1}^m \\text{diag}(X g_i > 0) X W_{1i} \\cdot
        W_{2i},

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons.
        bias: whether or not the model uses a bias term.
        G: the gate vectors for the Gated ReLU activation stored as a
            (d x p) matrix.
        G_bias: an optional vector of biases for the gates.
        parameters: the parameters of the model stored as a list of tensors.
    """

    def __init__(
        self,
        G: np.ndarray,
        c: int = 1,
        bias: bool = False,
        G_bias: Optional[np.ndarray] = None,
    ) -> None:
        """
        Args:
            G: (d x p) matrix of get vectors, where p is the number neurons.
            c: the output dimension.
            bias: whether or not to include a bias term.
            G_bias: a vector of bias parameters for the gates.
                Note that `bias` must be True for this to be supported.
        """
        super().__init__(G, c, bias, G_bias)

        # one linear model per gate vector
        if self.bias:
            self.parameters = [
                np.zeros((self.p, self.d)),  # first layer weights
                np.zeros((self.p)),  # first layer bias
                np.zeros((self.c, self.p)),  # second layer weights
            ]
        else:
            self.parameters = [
                np.zeros((self.p, self.d)),
                np.zeros((self.c, self.p)),
            ]

    def get_parameters(self) -> List[np.ndarray]:
        """Get the model parameters.

        Returns: A list of model parameters.
        """

        return self.parameters

    def set_parameters(self, parameters: List[np.ndarray]):
        """Set the model parameters.

        This method safety checks the dimensionality of the new parameters.

        Args:
            parameters: the new model parameters.
        """
        assert parameters[0].shape == (self.p, self.d)
        if self.bias:
            assert len(parameters) == 3
            assert parameters[1].shape == (self.p,)
            assert parameters[2].shape == (self.c, self.p)
        else:
            assert len(parameters) == 2
            assert parameters[1].shape == (self.c, self.p)

        self.parameters = parameters

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Compute the model predictions for a given dataset.

        Args:
            X: an (n,d) array containing the data examples on which to predict.

        Returns:
            - h(X): the model predictions for X.
        """
        D = super().compute_activations(X)

        Z = X @ self.parameters[0].T

        if self.bias:
            Z += self.parameters[1]

        return np.multiply(D, Z) @ self.parameters[-1].T


class ConvexReLU(GatedModel):
    """Convex reformulation of a ReLU Network with two-layers.

    This model has the prediction function

    .. math:: g(X) = \\sum_{D_i \\in \\mathcal{D}}^m D_i X (v_{i} - w_{i}),

    A one-vs-all strategy is used to extend the model to multi-dimensional
        targets.

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons.
        bias: whether or not the model uses a bias term.
        G: the gate vectors used to generate the activation patterns
            :math:`D_i`, stored as a (d x p) matrix.
        G_bias: an optional vector of biases for the gates.
        parameters: the parameters of the model stored as a list of two
            (c x p x d) matrices.
    """

    def __init__(
        self,
        G: np.ndarray,
        c: int = 1,
        bias: bool = False,
        G_bias: Optional[np.ndarray] = None,
    ) -> None:
        """Construct a new convex Gated ReLU model.

        Args:
            G: (d x p) matrix of get vectors, where p is the number neurons.
            c: the output dimension.
            bias: whether or not to include a bias term.
            G_bias: a vector of bias parameters for the gates.
                Note that `bias` must be True for this to be supported.
        """

        super().__init__(G, c, bias, G_bias)

        # one linear model per gate vector
        if self.bias:
            self.parameters = [
                np.zeros((c, self.p, self.d)),
                np.zeros((c, self.p)),
                np.zeros((c, self.p, self.d)),
                np.zeros((c, self.p)),
            ]
        else:
            self.parameters = [
                np.zeros((c, self.p, self.d)),
                np.zeros((c, self.p, self.d)),
            ]

    def get_parameters(self) -> List[np.ndarray]:
        """Get the model parameters."""
        return self.parameters

    def set_parameters(self, parameters: List[np.ndarray]):
        """Set the model parameters.

        This method safety checks the dimensionality of the new parameters.

        Args:
            parameters: the new model parameters.
        """
        if self.bias:
            assert len(parameters) == 4
            assert parameters[0].shape == (self.c, self.p, self.d)
            assert parameters[2].shape == (self.c, self.p)
            assert parameters[3].shape == (self.c, self.p, self.d)
            assert parameters[4].shape == (self.c, self.p)
        else:
            assert len(parameters) == 2
            assert parameters[0].shape == (self.c, self.p, self.d)
            assert parameters[1].shape == (self.c, self.p, self.d)

        self.parameters = parameters

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Compute the model predictions for a given dataset.

        Args:
            X: an (n,d) array containing the data examples on which to predict.

        Returns:
            - g(X): the model predictions for X.
        """
        D = super().compute_activations(X)

        p_diff = self.parameters[0] - self.parameters[1]

        if self.bias:
            bias_diff = self.parameters[1] - self.parameters[3]
            p_diff = self.parameters[0] - self.parameters[2]

            Z = np.einsum("ij, lkj->lik", X, p_diff) + bias_diff

            # TODO: need to check this computation.
            return np.einsum("lik, ik->il", Z, D)
        else:
            p_diff = self.parameters[0] - self.parameters[1]
            return np.einsum("ij, lkj, ik->il", X, p_diff, D)


class NonConvexReLU(Model):
    """Convex reformulation of a ReLU Network with two-layers.

    This model has the prediction function

    .. math:: h(X) = \\sum_{i=1}^p (X W_{1i}^{\\top})_+ \\cdot W_{2i}^{\\top},

    Attributes:
        c: the output dimension.
        d: the input dimension.
        p: the number of neurons.
        bias: whether or not the model uses a bias term.
        parameters: the parameters of the model stored as a list of matrices
            with shapes: [(p x d), (c x p)]
    """

    def __init__(
        self,
        d: int,
        p: int,
        c: int = 1,
        bias: bool = False,
    ) -> None:
        """Construct a new convex Gated ReLU model.

        Args:
            d: the input dimension.
            p: the number of neurons or "hidden units" in the network.
            c: the output dimension.
            bias: whether or not to include a bias term.
        """

        self.d = d
        self.p = p
        self.c = c
        self.bias = bias

        if self.bias:
            self.parameters = [
                np.zeros((self.p, self.d)),  # first layer weights
                np.zeros((self.p)),  # first layer bias
                np.zeros((self.c, self.p)),  # second layer weights
            ]
        else:
            self.parameters = [
                np.zeros((self.p, self.d)),  # first layer weights
                np.zeros((self.c, self.p)),  # second layer weights
            ]

    def get_parameters(self) -> List[np.ndarray]:
        """Get the model parameters.

        Returns: list of model parameters.
        """

        return self.parameters

    def set_parameters(self, parameters: List[np.ndarray]):
        """Set the model parameters.

        This method safety checks the dimensionality of the new parameters.

        Args:
            parameters: the new model parameters.
        """
        if self.bias:
            assert len(parameters) == 3
            assert parameters[0].shape == (self.p, self.d)
            assert parameters[1].shape == (self.p,)
            assert parameters[2].shape == (self.c, self.p)
        else:
            assert len(parameters) == 2
            assert parameters[0].shape == (self.p, self.d)
            assert parameters[1].shape == (self.c, self.p)

        self.parameters = parameters

    def __call__(self, X: np.ndarray) -> np.ndarray:
        """Compute the model predictions for a given dataset.

        Args:
            X: an (n,d) array containing the data examples on which to predict.

        Returns:
            - h(X): the model predictions for X.
        """
        Z = X @ self.parameters[0].T

        if self.bias:
            Z += self.parameters[1]

        return np.max(Z, 0) @ self.parameters[-1].T



"""Convert solvers from :module:`scnn.solvers` into internal optimizers
and optimization procedures."""

from typing import Optional

import scnn.private.prox as prox
from scnn.metrics import Metrics
from scnn.private.methods import (
    FISTA,
    AugmentedLagrangian,
    ConstrainedHeuristic,
    LagrangianGradNorm,
    CVXPYGatedReLUSolver,
    CVXPYReLUSolver,
    DoubleLoopProcedure,
    GradientNorm,
    IterativeOptimizationProcedure,
    LagrangianGradNorm,
    Lassplore,
    LinearSolver,
    MultiplicativeBacktracker,
    OptimizationProcedure,
)
from scnn.private.methods import Optimizer as InteralOpt
from scnn.private.methods import ProximalCleanup, QuadraticBound
from scnn.private.methods import ApproximateConeDecomposition as ACD
from scnn.regularizers import L1, L2, FeatureGL1, NeuronGL1, Regularizer
from scnn.solvers import (
    AL,
    RFISTA,
    ApproximateConeDecomposition,
    CVXPYSolver,
    LeastSquaresSolver,
    Optimizer,
)

from scnn.models import ConvexGatedReLU, ConvexReLU
from scnn.regularizers import NeuronGL1
from scnn.private.interface import build_internal_regularizer


def build_prox_operator(
    regularizer: Optional[Regularizer] = None,
) -> prox.ProximalOperator:
    """Convert public facing regularizer into proximal operator.

    Args:
        regularizer: a public-facing regularizer object.

    Returns:
        A proximal operator for the regularizer.
    """
    lam = 0.0
    op: prox.ProximalOperator

    if regularizer is not None:
        lam = regularizer.lam

    if isinstance(regularizer, L2):
        op = prox.L2(lam)
    elif isinstance(regularizer, L1):
        op = prox.L1(lam)
    elif isinstance(regularizer, NeuronGL1):
        op = prox.GroupL1(lam)
    elif isinstance(regularizer, FeatureGL1):
        op = prox.FeatureGroupL1(lam)
    elif regularizer is None:
        op = prox.Identity()
    else:
        raise ValueError(
            f"Optimizer does not support regularizer {regularizer}."
        )

    return op


def build_fista(
    regularizer: Optional[Regularizer],
) -> FISTA:
    """Helper function for constructing a default instance of R-FISTA.

    Args:
        regularizer: a public-facing regularizer object.

    Returns:
        Instance of the R-FISTA optimizer.
    """

    prox = build_prox_operator(regularizer)

    return FISTA(
        10.0,
        QuadraticBound(),
        MultiplicativeBacktracker(beta=0.8),
        Lassplore(alpha=1.25, threshold=5.0),
        prox=prox,
    )


def build_optimizer(
    optimizer: Optimizer,
    regularizer: Optional[Regularizer],
    metrics: Metrics,
) -> OptimizationProcedure:
    """Convert public facing solver into an internal optimization procedure.

    Args:
        optimizer: a public-facing optimizer object.
        regularizer: a public-facing regularizer object.
        metrics: a metrics object specifying which metrics to collect during optimization.

    Returns:
        An optimization procedure.
    """

    opt: InteralOpt
    opt_proc: OptimizationProcedure

    if isinstance(optimizer, RFISTA):

        max_iters = optimizer.max_iters
        term_criterion = GradientNorm(optimizer.tol)
        opt = build_fista(regularizer)

        opt_proc = IterativeOptimizationProcedure(
            opt,
            max_iters,
            term_criterion,
            name="fista",
            divergence_check=True,
            log_freq=metrics.metric_freq,
        )

    elif isinstance(optimizer, AL):

        inner_term_criterion = GradientNorm(optimizer.tol)
        outer_term_criterion = ConstrainedHeuristic(
            optimizer.tol, optimizer.constraint_tol
        )
        # outer_term_criterion = LagrangianGradNorm(optimizer.tol)

        sub_opt = build_fista(regularizer)
        opt = AugmentedLagrangian(
            use_delta_init=True,
            subprob_tol=optimizer.tol,
        )
        opt_proc = DoubleLoopProcedure(
            sub_opt,
            opt,
            optimizer.max_primal_iters,
            optimizer.max_dual_iters,
            inner_term_criterion,
            outer_term_criterion,
            max_total_iters=optimizer.max_primal_iters,
            name="al",
            divergence_check=False,
            log_freq=metrics.metric_freq,
        )
    elif isinstance(optimizer, LeastSquaresSolver):
        if not (regularizer is None or isinstance(regularizer, L2)):
            raise ValueError(
                "LeastSquaresSolver only supports L2-regularization, or no regularizer."
            )
        linear_solver = LinearSolver(
            optimizer.solver, optimizer.max_iters, optimizer.tol, None
        )

        opt_proc = OptimizationProcedure(linear_solver)

    elif isinstance(optimizer, CVXPYSolver):
        # check which formulation we need to solve.
        model = optimizer.model

        post_process = None
        if optimizer.clean_sol:
            prox = build_prox_operator(regularizer)
            post_process = ProximalCleanup(prox)

        if isinstance(model, ConvexReLU):
            opt = CVXPYReLUSolver(optimizer.solver, optimizer.solver_kwargs)
        elif isinstance(model, ConvexGatedReLU):
            opt = CVXPYGatedReLUSolver(
                optimizer.solver, optimizer.solver_kwargs
            )
        else:
            raise ValueError(f"Model {model} not recognized by CVXPYSolver.")

        opt_proc = OptimizationProcedure(opt, post_process=post_process)

    elif isinstance(optimizer, ApproximateConeDecomposition):
        decomp_regularizer = NeuronGL1(optimizer.rho)
        int_decomp_reg = build_internal_regularizer(decomp_regularizer)
        int_decomp_prox = build_prox_operator(decomp_regularizer)
        opt_proc = build_optimizer(optimizer.rfista, regularizer, metrics)

        post_opt = ACD(
            int_decomp_reg,
            int_decomp_prox,
            optimizer.d_max_iters,
            optimizer.d_tol,
        )

        opt_proc.post_process = post_opt

        return opt_proc
    else:
        raise ValueError(f"Optimizer object {optimizer} not supported.")

    return opt_proc



"""End-to-end tests for regularizers.

TODO:
    - Test solving ReLU and Gated ReLU problems with different regularizers and with appropriate solvers.
    - Should we solve both the ReLU and Gated ReLU problems or just the ReLU?
    - L2-squared and L1.
"""

import unittest

from scnn.private.utils.data.synthetic import gen_regression_data
from scnn.regularizers import NeuronGL1, FeatureGL1, L2, L1
from scnn.optimize import optimize


class TestRegularizers(unittest.TestCase):
    """Test solving convex reformulations with different regularizers."""

    def setUp(self):
        # Generate realizable synthetic classification problem (ie. Figure 1)
        n_train = 100
        n_test = 100
        d = 25
        c = 2
        kappa = 1
        self.max_neurons = 100

        (
            (self.X_train, self.y_train),
            (self.X_test, self.y_test),
            _,
        ) = gen_regression_data(123, n_train, n_test, d, c, kappa=kappa)
        self.lam = 0.0001

    def test_l2_squared(self):
        """Test Gated ReLU with L2 regularization."""
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            L2(self.lam),
        )

    def test_l1(self):
        """Test Gated ReLU with L1 regularization."""
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            L1(self.lam),
        )

    def test_neuron_gl1(self):
        """Test Gated ReLU with neuron-wise Group-L1 regularization."""
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            NeuronGL1(self.lam),
        )

    def test_feature_gl1(self):
        """Test Gated ReLU with feature-wise Group-L1 regularization."""
        cvx_model, metrics = optimize(
            "gated_relu",
            self.max_neurons,
            self.X_train,
            self.y_train,
            self.X_test,
            self.y_test,
            FeatureGL1(self.lam),
        )


if __name__ == "__main__":
    unittest.main()



"""Optimization procedure for iterative methods (e.g. gradient descent)."""
from logging import Logger, root, INFO
from typing import Dict, Any, List, Tuple, Callable, Optional
import timeit

from tqdm.auto import tqdm  # type: ignore

import lab

from scnn.private.methods.optimization_procedures.optimization_procedure import (
    ITER_LOG_FREQ,
    METRIC_FREQ,
)
from scnn.private.methods.optimization_procedures.iterative import (
    IterativeOptimizationProcedure,
)

from scnn.private.methods.optimizers import (
    Optimizer,
    MetaOptimizer,
    ProximalOptimizer,
)
from scnn.private.models.model import Model
from scnn.private.metrics import (
    update_metrics,
    init_metrics,
    format_recent_metrics,
)
from scnn.private.methods.termination_criteria import TerminationCriterion


class DoubleLoopProcedure(IterativeOptimizationProcedure):

    """An iterative optimization procedure."""

    optimizer: Optimizer

    def __init__(
        self,
        inner_optimizer: Optimizer,
        outer_optimizer: MetaOptimizer,
        inner_max_iters: int,
        outer_max_iters: int,
        inner_term_criterion: TerminationCriterion,
        outer_term_criterion: TerminationCriterion,
        name: str = "",
        divergence_check: bool = False,
        batch_size: Optional[int] = None,
        log_freq: int = ITER_LOG_FREQ,
        metric_freq: int = METRIC_FREQ,
        pre_process: Optional[Callable] = None,
        post_process: Optional[Callable] = None,
        max_total_iters: Optional[int] = None,
    ):
        """
        :param inner_optimizer: an optimizer instance that will be used to execute the inner
            optimization loop.
        :param outer_optimizer: an optimizer instance that will be used to execute the outer
            optimization loop.
        :param inner_max_iters: the maximum number of iterations to run inner optimizer per
            outer step.
        :param outer_max_iters: the maximum number of iterations to run the outer optimizer.
        :param inner_term_criterion: the criterion to use when checking for early termination
            of the inner optimization routine.
        :param outer_term_criterion: the criterion to use when checking for early termination
            of the outer optimization routine.
        :param name: an optional name to display when printing the progress bar.
        :param divergence_check: an whether or not the optimization procedure should
            check for divergence behavior and terminate early.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :param log_freq: the frequency at which to information during optimization.
        :param metric_freq: the frequency at which to collect metrics, like objective,
            gradient-norm, etc, during optimization.
        :param pre_process: (optional) a function to call on the model after the it is
            initialized but *before* optimization starts.
        :param pre_process: (optional) a function to call on the model *after* optimization
            is complete.
        :param max_total_iters: the maximum number of inner steps that are permitted.
        """
        self.inner_optimizer = inner_optimizer
        self.outer_optimizer = outer_optimizer
        self.inner_max_iters = inner_max_iters
        self.outer_max_iters = outer_max_iters
        self.inner_term_criterion = inner_term_criterion
        self.outer_term_criterion = outer_term_criterion

        self.name = name
        self.divergence_check = divergence_check
        self.batch_size = batch_size

        self.log_freq = log_freq
        self.metric_freq = metric_freq

        self.pre_process = pre_process
        self.post_process = post_process
        self.max_total_iters = max_total_iters

    def reset(self):
        """Reset the optimization procedure (including any attached optimizers)
        to their original state."""
        self.outer_optimizer.reset()
        self.inner_optimizer.reset()

    def __call__(
        self,
        logger: Logger,
        model: Model,
        initializer: Callable[[Model], Model],
        train_data: Tuple[lab.Tensor, lab.Tensor],
        test_data: Tuple[lab.Tensor, lab.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]] = None,
        callback: Callable = None,
    ) -> Tuple[Dict[str, Any], Model, Dict[str, Any]]:
        """Optimize a model using an iterative optimize method by running a
        "optimization" loop. The loop runs for 'max_iters' number of
        iterations, collects optional metrics at every iteration, and will
        terminate early if the conditions of 'term_criterion' are met.

        :param logger: a logging.Logger object that can be used to log information during
            optimization.
        :param model: the model to optimize.
        :param initializer: a function that can be called to initialize the model.
        :param train_data: an (X, y) tuple representing the training set.
        :param test_data: an (X_test, y_test) tuple representing the test data.
        :param metrics: a tuple of the form (train_metrics, test_metrics, additional_metrics)
            specifying the metrics to be computed on the training set, test set, and data-independent
            metrics.
        :returns: (exit_status, model, metrics) --- execution information, the optimized model, and metrics from optimization.
        """
        # pre-optimization setup
        (model, metrics_log, exit_status, objective, grad,) = self._pre_optimization(
            logger,
            model,
            initializer,
            train_data,
            test_data,
            metrics,
            final_metrics,
            self.inner_optimizer.step_size,
        )
        (X, y), f0, start_time = train_data, objective, None
        # training loop
        verbose = root.level <= INFO

        # whether or not to terminate the outer optimization procedure.
        terminate = False
        total_itrs = 0

        # outer loop
        for outer_itr in tqdm(
            range(self.outer_max_iters),
            desc="Outer " + self.name,
            disable=(not verbose),
        ):

            # inner loop says to terminate optimization.
            if terminate:
                break

            # capture time-cost of outer optimizer.

            start_time = self._get_start_time(start_time)
            # outer update
            (
                model,
                self.inner_term_criterion,
                self.inner_optimizer,
                sp_exit_state,
            ) = self.outer_optimizer.step(
                model, self.inner_term_criterion, self.inner_optimizer, X, y
            )

            # compute objective and gradient after outer update.
            objective = model.objective(X, y, batch_size=self.batch_size)
            grad = model.grad(
                X,
                y,
                batch_size=self.batch_size,
                step_size=self.inner_optimizer.step_size,
            )

            # inner loop
            for inner_itr in tqdm(
                range(self.inner_max_iters),
                desc="Inner " + self.name,
                disable=(not verbose),
            ):

                if (
                    self.max_total_iters is not None
                    and total_itrs >= self.max_total_iters
                ):
                    terminate = True
                    break
                total_itrs += 1

                start_time = self._get_start_time(start_time)

                if inner_itr % self.log_freq == 0 and verbose:
                    tqdm.write(format_recent_metrics(metrics_log, metrics))

                # outer termination criterion
                if self.outer_term_criterion(model, X, y, objective, grad):
                    exit_status["success"] = True
                    logger.info(
                        f"*Outer* termination criterion satisfied at iteration {outer_itr}/{self.outer_max_iters}. Exiting *outer* optimization loop."
                    )
                    terminate = True
                    break

                # inner termination criterion
                model = self.outer_optimizer.inner(model)
                if self.inner_term_criterion(model, X, y, objective, grad):
                    exit_status["success"] = True
                    logger.info(
                        f"*Inner* termination criterion satisfied at iteration {inner_itr}/{self.inner_max_iters}. Exiting *inner* optimization loop."
                    )
                    break

                # check for divergence
                if self.divergence_check and objective > 1000 * f0:
                    exit_status["success"] = False
                    logger.warning(
                        f"Method diverged at {inner_itr}/{self.inner_max_iters}. Exiting optimization loop."
                    )
                    break

                model, objective, sp_exit_state = self.inner_optimizer.step(
                    model, X, y, objective, grad, self.batch_size
                )

                if callback is not None:
                    model = callback(model, X, y)

                model = self.outer_optimizer.outer(model)

                # compute objective and gradient for next iteration.
                if objective is None:
                    objective = model.objective(X, y, batch_size=self.batch_size)

                grad = model.grad(
                    X,
                    y,
                    batch_size=self.batch_size,
                    step_size=self.inner_optimizer.step_size,
                )

                # calculate time and other metrics.
                if inner_itr % self.metric_freq == 0:
                    metrics_log, start_time = self._record_time(metrics_log, start_time)
                    metrics_log = update_metrics(
                        metrics_log,
                        model,
                        sp_exit_state,
                        train_data,
                        test_data,
                        metrics,
                        objective,
                        grad,
                        batch_size=self.batch_size,
                    )

            if inner_itr == self.inner_max_iters - 1:
                logger.warning(
                    "Max iterations reached before *inner* termination criterion was satisfied. Exiting *inner* optimization loop."
                )

        # post-optimization clean-up.
        model, exit_status, metrics_log = self._post_optimization(
            logger,
            model,
            train_data,
            test_data,
            metrics,
            final_metrics,
            exit_status,
            metrics_log,
            start_time,
            self.inner_optimizer.step_size,
        )
        # collect final run-information
        if outer_itr == self.outer_max_iters - 1 or total_itrs == self.max_total_iters:
            exit_status["success"] = False
            logger.warning(
                "Max iterations reached before *outer* termination criterion was satisfied. Exiting *outer* optimization loop."
            )
        exit_status["outer_iterations"] = outer_itr + 1
        exit_status["total_inner_iterations"] = total_itrs

        return exit_status, model, metrics_log



"""Base class for proximal-gradient optimizers."""
from typing import Callable, Optional, Tuple, Any, Dict

import lab

from scnn.private.models.model import Model
from scnn.private.methods.optimizers.optimizer import Optimizer, LSOptimizer
from scnn.private.prox import ProximalOperator
from scnn.private.methods.line_search import (
    StepSizeUpdater,
    LSCondition,
    Backtracker,
)


class ProximalOptimizer(Optimizer):

    """The base class for proximal-gradient-type methods.

    Optimizer is a stateful wrapper around a 'step_fn' --- a function which can
    be called to compute one step of an iterative optimization method.
    """

    def __init__(
        self,
        step_fn: Callable,
        step_size: float,
        prox: ProximalOperator,
        update_step_size: StepSizeUpdater,
    ):

        """
        :param step_fn: a function to call to execute one step of the iterative
            method. It should have the same signature as 'core.proximal_gradient.proximal_gradient_step'.
        :param step_size: the constant step-size to use.
        :param prox: a proximal operator. See 'proximal_ops'.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        super().__init__(step_fn, step_size, update_step_size)
        self.prox = prox

    def step(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        f0: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
        batch_index: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> Tuple[Model, Optional[float], Dict[str, Any]]:
        """Execute one step of an iterative optimization method.

        :param model: the model to update.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param f0: (optional) current objective value. Can be provided to re-use computations.
        :param grad: (optional) current objective gradient. Can be provided to re-use computations.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the updated model and empty exit state (no sub-procedures are called).
        """
        # compute gradient without non-smooth regularizer.
        grad = model.grad(X, y, ignore_regularizer=True, batch_size=batch_size)

        old_weights = model.weights
        model.weights = self.step_fn(model.weights, grad, self.step_size, self.prox)

        # update step-size
        self.step_size = self.update_step_size(
            self.step_size,
            self.old_step_size,
            model.weights - old_weights,
            self.f0,
            self.f1,
            grad,
        )

        return model, None, {"step_size": self.step_size}


class ProximalLSOptimizer(LSOptimizer):

    """First-order optimization method with line-search.

    Like optimizer, this is a stateful wrapper around a 'step_fn', which can be
    called to execute one step of the f.o. method, including the line-search.
    """

    def __init__(
        self,
        step_fn: Callable,
        init_step_size: float,
        ls_cond: LSCondition,
        backtrack_fn: Backtracker,
        update_step_size: StepSizeUpdater,
        prox: ProximalOperator,
    ):
        """
        :param step_fn: a function to call to execute one step of the iterative
            method. It should have the same signature as 'methods.gradient_descent.gd_ls'.
        :param init_step_size: first step-size to try when running the line-search.
        :param ls_cond: the line-search condition to check.
        :param backtrack: a rule for calculating the next step-size to try.
        :param update_step_size: a rule for updating the step-size after each line-search.
        :param prox: a proximal operator. See 'proximal_ops'.
        """
        super().__init__(
            step_fn, init_step_size, ls_cond, backtrack_fn, update_step_size
        )
        self.prox = prox

    def step(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        f0: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
        batch_index: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> Tuple[Model, Optional[float], Dict[str, Any]]:
        """Execute one step of an iterative optimization method.

        :param model: the model to update.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param f0: (optional) current objective value. Can be provided to re-use computations.
        :param grad: (optional) current objective gradient. Can be provided to re-use computations.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the updated model and exit state of the line-search.
        """
        # model must be wrapped in a regularizer.
        obj_fn, grad_fn = model.get_closures(
            X, y, ignore_regularizer=True, batch_size=batch_size
        )

        grad = grad_fn(model.weights)

        if self.f1 is None:
            self.f0 = obj_fn(model.weights)
        else:
            self.f0 = self.f1

        old_weights = model.weights
        model.weights, self.f1, self.step_size, exit_state = self.step_fn(
            model.weights,
            self.f0,
            grad,
            grad,
            obj_fn,
            grad_fn,
            self.step_size,
            self.ls_cond,
            self.backtrack_fn,
            prox=self.prox,
        )

        # update step-size
        self.step_size = self.update_step_size(
            self.step_size,
            self.old_step_size,
            model.weights - old_weights,
            self.f0,
            self.f1,
            grad,
        )
        self.old_step_size = self.step_size

        return model, None, exit_state



"""Rules for backtracking and updating step-sizes during a line-search."""
from typing import List, Optional

import lab


class Backtracker:

    """Base class for backtracking conditions."""

    def __call__(
        self,
        step_size: float,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Pick the next candidate step-size for a line-search.

        :param step_size: the step_size previously tried.
        :returns: a new step-size to try.
        """
        raise NotImplementedError("Backtracking methods must implement '__call__'!")


class MultiplicativeBacktracker(Backtracker):

    """Step-size selection by simple multiplicative backtracking."""

    def __init__(self, beta=0.9):
        """
        :param beta: the backtracking parameter.
        """
        self.beta = beta

    def __call__(
        self,
        step_size: float,
        grad: Optional[lab.Tensor] = None,
    ) -> float:
        """Step-size selection by simple backtracking,

            step_size <- step_size * beta.
        :param step_size: the step_size previously tried.
        :returns: a new step-size to try.
        """

        return step_size * self.beta



"""Line-search."""

# ===== module exports ===== #

from .backtrack import Backtracker, MultiplicativeBacktracker
from .conditions import (
    LSCondition,
    FSS,
    QuadraticBound,
    DiagQB,
    Armijo,
)
from .step_size_updates import (
    StepSizeUpdater,
    KeepNew,
    KeepOld,
    ForwardTrack,
    Lassplore,
)

__all__ = [
    "Backtracker",
    "MultiplicativeBacktracker",
    "LSCondition",
    "FSS",
    "QuadraticBound",
    "DiagQB",
    "Armijo",
    "StepSizeUpdater",
    "KeepNew",
    "KeepOld",
    "ForwardTrack",
    "Lassplore",
]



"""Proximal gradient descent with line-search."""

from typing import Callable, Tuple, Dict, Any
from functools import partial

import numpy as np

import lab

from scnn.private.methods.core import ls
from scnn.private.methods.line_search import Backtracker, LSCondition
from scnn.private.prox import ProximalOperator

# step operators


def proximal_gradient_step(
    w: lab.Tensor,
    descent_dir: lab.Tensor,
    step_size: float,
    prox: ProximalOperator,
) -> lab.Tensor:
    """Take one step of proximal gradient descent.

    :param w: the parameters to be updated.
    :param grad: the gradient of the smooth component of the loss function.
    :param step_size: the step-size to use.
    :param prox: the proximal operator. It must take a step-size parameter.
    :returns: updated parameters
    """

    w_plus = w - step_size * descent_dir
    return prox(w_plus, step_size)


def fista_step(
    w: lab.Tensor,
    descent_dir: lab.Tensor,
    step_size: float,
    prox: ProximalOperator,
    v: lab.Tensor,
    t: float,
) -> Tuple[lab.Tensor, lab.Tensor, float]:
    """
    :param w: the parameters to be updated.
    :param grad: the gradient of the smooth component of the loss function.
    :param step_size: the step-size to use.
    :param prox: the proximal operator. It must take a step-size parameter.
    :param v: the extrapolation sequence.
    :param t: the extrapolation parameter.
    :returns: (w_plus, v_plus, t_plus) --- the updated sequences.
    """

    w_plus = proximal_gradient_step(v, descent_dir, step_size, prox)
    t_plus = 1 + np.sqrt(1 + 4 * t ** 2) / 2
    v_plus = w_plus + (t - 1) * (w_plus - w) / t_plus

    return w_plus, v_plus, t_plus


# main method: proximal gradient descent with line-search.


def proximal_gradient_ls(
    w: lab.Tensor,
    f0: float,
    descent_dir: lab.Tensor,
    grad: lab.Tensor,
    obj_fn: Callable,
    grad_fn: Callable,
    init_step_size: float,
    ls_cond: LSCondition,
    backtrack: Backtracker,
    prox: ProximalOperator,
) -> Tuple[lab.Tensor, float, float, Dict[str, Any]]:
    """Take one step of proximal gradient descent using a line-search to pick
    the step-size.

    :param w: the parameters to be updated.
    :param f0: the objective function evaluated at w.
    :param descent_dir: the descent direction for the step.
    :param grad: the gradient of the smooth component of the loss function.
    :param obj_fn: function that returns the objective when called at w.
    :param grad_fn: function that returns the gradient of the smooth component of the loss function.
    :param init_step_size: the first step-size to use.
    :param ls_cond: the line-search condition to check.
    :param backtrack: a rule for calculating the next step-size to try.
    :param prox: the proximal operator. It must take a step-size parameter.
    :returns: (w_next, f1, step_size, exit_state): the updated parameters, objective value, new step-size, and exit state.
    """

    # input the proximal operator and extra arguments.
    test_fn = partial(proximal_gradient_step, prox=prox)

    # line-search *through* the proximal operator.
    return ls(
        test_fn,
        w,
        f0,
        descent_dir,
        grad,
        obj_fn,
        grad_fn,
        init_step_size,
        ls_cond,
        backtrack,
    )


def fista_ls(
    w: lab.Tensor,
    f0: float,
    descent_dir: lab.Tensor,
    grad: lab.Tensor,
    obj_fn: Callable,
    grad_fn: Callable,
    init_step_size: float,
    ls_cond: LSCondition,
    backtrack: Backtracker,
    prox: ProximalOperator,
    v: lab.Tensor,
    t: float,
    mu: float = 0.0,
) -> Tuple[lab.Tensor, lab.Tensor, float, float, float, Dict[str, Any]]:
    """Take one step of proximal gradient descent using a line-search to pick
    the step-size.

    :param w: the parameters to be updated.
    :param f0: the objective function evaluated at w.
    :param descent_dir: the descent direction for the step.
    :param grad: the gradient of the smooth component of the loss function.
    :param obj_fn: function that returns the objective when called at w.
    :param grad_fn: function that returns the gradient of the smooth component of the loss function.
    :param init_step_size: the first step-size to use.
    :param ls_cond: the line-search condition to check.
    :param backtrack: a rule for calculating the next step-size to try.
    :param prox: the proximal operator. It must take a step-size parameter.
    :param v: the extrapolation sequence.
    :param t: the extrapolation parameter.
    :param mu: (optional) a lower-bound on the strong-convexity parameter of the objective.
        The method defaults to the parameter sequence for non-strongly convex functions when
        mu is not supplied.
    :returns: (w_next, f1, step_size, exit_state): the updated parameters, objective value, new step-size, and exit state.
    """

    # line-search starting from secondary (extrapolation) sequence.
    w_plus, f1, step_size, exit_state = proximal_gradient_ls(
        v,
        f0,
        descent_dir,
        grad,
        obj_fn,
        grad_fn,
        init_step_size,
        ls_cond,
        backtrack,
        prox,
    )
    # release memory
    t_plus = t

    if mu == 0.0:
        t_plus = 1 + np.sqrt(1 + 4 * t ** 2) / 2
        beta = (t - 1) / t_plus
    else:
        sqrt_kappa = np.sqrt(1 / (step_size * mu))
        beta = (sqrt_kappa - 1) / (sqrt_kappa + 1)

    v = w_plus + beta * (w_plus - w)

    return w_plus, v, t_plus, f1, step_size, exit_state



"""One-vs-all-classification model."""

from typing import Optional, Tuple, List

import lab

from scnn.private.models.model import Model
from scnn.private.models.regularizers import Regularizer
from scnn.private.loss_functions import squared_error


class OneVsAllModel(Model):
    """A one-vs-all model."""

    def __init__(
        self,
        d: int,
        per_class_models: List[Model],
        c: int = 1,
        regularizer: Optional[Regularizer] = None,
    ):
        assert len(per_class_models) == c
        super().__init__(regularizer)

        self.per_class_models = per_class_models

        self.d = d
        self.c = c

    @property
    def weights(self):
        weights = lab.stack([model.weights for model in self.per_class_models])
        return weights

    def _forward(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: (NOT USED) parameter at which to compute the forward pass.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
            Defaults to self.D or manual computation depending on the value of self._train.
        :returns: predictions for X.
        """
        return lab.concatenate(
            [model(X) for model in self.per_class_models], axis=1
        )

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n) array containing the data targets.
        :param w: (NOT USED) specific parameter at which to compute the forward pass.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: objective L(f, (X, y)).
        """
        return squared_error(self._forward(X, None), y) / (
            2 * self._scaling(y, scaling)
        )

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        flatten: bool = False,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the l2 objective with respect to the model
        weights.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: (NOT USED) parameter at which to compute the gradient.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
            Defaults to self.D or manual computation depending on the value of self._train.
        :param flatten: whether or not to flatten the blocks of the gradient into a single vector.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        grad = lab.stack(
            [
                model.grad(X, y, w=None, flatten=False, scaling=scaling)
                for model in self.per_class_models
            ]
        )

        if flatten:
            grad = grad.reshape(-1)

        return grad



"""Basic interface for CVXPY-base solvers."""

from typing import Dict, Any, Tuple

import cvxpy as cp
import numpy as np

from scnn.private.methods.external_solver import ExternalSolver
from scnn.private.models import Model


class CVXPYSolver(ExternalSolver):
    """Interface for solvers based on the `CVXPY <https://www.cvxpy.org>`_ DSL.

    Attributes:
        solver: underlying solver to use with CVXPY.
        solver_kwargs: a dictionary of keyword arguments to be passed directly to the underlying solver.
    """

    def __init__(self, solver: str = "ecos", solver_kwargs={}):
        """Initialize the CVXPY-based solver.

        Args:
            solver: a string identifying the solver to use with CVXPY.
                We only support 'ecos', 'cvxopt', 'scs', 'gurobi', and 'mosek' at the moment.
        """

        self.solver_kwargs = solver_kwargs

        # save the desired solver
        if solver == "ecos":
            self.solver = cp.ECOS_BB
        elif solver == "cvxopt":
            self.solver = cp.CVXOPT
        elif solver == "scs":
            self.solver = cp.SCS
        # note: the following are commercial solvers requiring a licence.
        elif solver == "gurobi":
            self.solver = cp.GUROBI
        elif solver == "mosek":
            self.solver = cp.MOSEK
        else:
            raise ValueError(f"CVXPY solver {solver} not recognized!")

    def __call__(
        self, model: Model, X: np.ndarray, y: np.ndarray
    ) -> Tuple[Model, Dict[str, Any]]:
        """Call the CVXPY solver to fit the model."""

        raise NotImplementedError("A CVXPY-based solver must implement '__call__'.")



"""
Pre- and post-processing functions to be run before and after optimization, respectively.
"""

from typing import Optional

import lab

from scnn.private.models import Model


class ProcessingFunction:

    """Abstract class for pre- and post-processing functions."""

    def __call__(self, model: Model, X: lab.Tensor, y: lab.Tensor) -> Model:
        """Prepare/update/refine a model before or after optimization
        begins/ends, optionally using the training set (X, y).

        :param model: the model instance to process.
        :param X: the training features.
        :param y: the training targets.
        :returns: processing_fn(model) --- the updated model.
        """
        raise NotImplementedError(
            "A pre- or post-processing function must implement `__call__`."
        )



"""Implementation of augmented Lagrangian method for fitting two-layer MLPs
with orthant constraints."""
from typing import Dict, Any, Tuple, Optional
import lab

from scnn.private.models.model import Model
from scnn.private.models import AL_MLP
from scnn.private.methods.optimizers import Optimizer
from scnn.private.methods.optimizers import MetaOptimizer
from scnn.private.methods.termination_criteria import TerminationCriterion
from scnn.private.methods.core import update_multipliers


class AugmentedLagrangian(MetaOptimizer):

    """Augmented Lagrangian method."""

    def __init__(
        self,
        use_delta_init: bool = True,
        subprob_tol: float = 1e-6,
        omega: float = 1e-3,
        eta_upper: float = 1e-2,
        eta_lower: float = 1e-3,
        tau: float = 2,
    ):
        """
        :param use_delta_init: whether or not to initialize delta using a heuristic rule that seeks to put the penalty
            on the same order as the initial function values.
        :param use_dynamic_tols: whether or not to update the tolerances dynamically during optimization.
            If true, the algorithm uses a Lancelot-like scheme to update delta, eta, and omega.
        :param init_scaling: multiplicate scaling for the initial delta value. Only matters when using the heuristic
            rule to initialize delta.
        :param omega: the starting tolerance to which each sub-problem will be solved. Will be updated dynamically
            during execution when using dynamic tolerances.
        :param eta: the starting tolerance for violation of the constraints. Only matters when using dynamic
            tolerances and will be changed dynamically during optimization.
        :param tau: the multiplicative increase for the penalty parameter when insufficient progress is made
            on the constraints. Only matters when using dynamic tolerances.
        :param c: the power which controls how the constraint tolerance evolves after successful iterations.
            Only matters when using dynamic tolerances.
        :param min_omega: the minimum value of the sub-problem optimization tolerance.
        """

        self.use_delta_init = use_delta_init

        self.subprob_tol = subprob_tol
        self.omega = omega
        self.eta_lower = eta_lower
        self.eta_upper = eta_upper
        self.tau = tau

        self.delta: Optional[float] = None

    def reset(self):
        """Reset the optimizer to its initial state."""

        self.delta = None

    def step(
        self,
        model: AL_MLP,
        inner_term_criterion: TerminationCriterion,
        inner_optimizer: Optimizer,
        X: lab.Tensor,
        y: lab.Tensor,
        batch_index: Optional[int] = None,
    ) -> Tuple[Model, TerminationCriterion, Optimizer, Dict[str, Any]]:
        """Execute one step of the augmented Lagrangian method. Warning: this
        will solve a another optimization problem as sub-routine, which can be
        costly.

        :param model: the model to update.
        :param inner_term_criterion: the termination criterion of the inner optimization
            routine to which this meta optimizer is being applied.
        :param inner_optimizer: the optimizer for the inner optimization
            routine to which this meta optimizer is being applied.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :returns: the updated model and exit status of any sub-procedures (ie. line-search)
        """

        # initial setup; no dual updates.
        if self.delta is None:
            self.delta = model.delta

            if self.use_delta_init:
                inner_term_criterion.tol = self.omega
                self.initializing_delta = True
            else:
                inner_term_criterion.tol = self.subprob_tol
                self.initializing_delta = False
        else:
            e_gap, i_gap = model.constraint_gaps(X)
            gap_norm = lab.sum(e_gap**2) + lab.sum(i_gap**2)

            # TODO: sometimes the lower window leads to weird failure cases.
            if self.initializing_delta:

                # the previous sub-problem did not make enough progress on the constraints;
                # increase penalty strength
                if gap_norm > self.eta_upper:
                    self.delta = self.delta * self.tau
                    model.delta = self.delta
                # the previous sub-problem made too much progress on the constraints;
                # decrease penalty strength
                elif gap_norm < self.eta_lower:
                    self.delta = self.delta / self.tau
                    model.delta = self.delta
                    inner_term_criterion.tol = (
                        inner_term_criterion.tol / self.tau
                    )
                else:
                    # gap within starting "window".
                    self.initializing_delta = False
                    inner_term_criterion.tol = self.subprob_tol

            else:
                # update the dual parameters.
                model.e_multipliers, model.i_multipliers = update_multipliers(
                    model.e_multipliers,
                    model.i_multipliers,
                    e_gap,
                    i_gap,
                    model.delta,
                )

                # mechanism for tightening solution tolerances.
                if self.tau * gap_norm < inner_term_criterion.tol:
                    # reduce tolerance to ensure continued progress.
                    inner_term_criterion.tol = (
                        inner_term_criterion.tol / self.tau
                    )

        inner_optimizer.reset()
        exit_status: Dict[str, Any] = {}

        return model, inner_term_criterion, inner_optimizer, exit_status



"""Convert models from :module:`scnn.models` into internal representations
and vice versa."""

from typing import Optional, List, Tuple

import numpy as np

import lab

from scnn.regularizers import (
    Regularizer,
    NeuronGL1,
    FeatureGL1,
    L2,
    L1,
)

from scnn.models import (
    Model,
    LinearModel,
    ConvexGatedReLU,
    NonConvexGatedReLU,
    ConvexReLU,
    NonConvexReLU,
)

from scnn.private.models import (
    ConvexMLP,
    AL_MLP,
    ReLUMLP,
    GatedReLUMLP,
    GroupL1Regularizer,
    FeatureGroupL1Regularizer,
    L2Regularizer,
    L1Regularizer,
    LinearRegression,
)
from scnn.activations import compute_activation_patterns

from scnn.private.models import Model as InternalModel
from scnn.private.models import Regularizer as InternalRegularizer


def build_internal_regularizer(
    regularizer: Optional[Regularizer] = None,
) -> InternalRegularizer:
    """Convert public-facing regularizer objects into private implementations.

    Args:
        regularizer: a regularizer object from the public API.

    Returns:
        An internal regularizer object with the same state as the public regularizer.
    """
    reg: Optional[InternalRegularizer] = None

    lam = 0.0
    if regularizer is not None:
        lam = regularizer.lam

    if isinstance(regularizer, NeuronGL1):
        reg = GroupL1Regularizer(lam)
    elif isinstance(regularizer, FeatureGL1):
        reg = FeatureGroupL1Regularizer(lam)
    elif isinstance(regularizer, L2):
        reg = L2Regularizer(lam)
    elif isinstance(regularizer, L1):
        reg = L1Regularizer(lam)

    return reg


def build_internal_model(
    model: Model, regularizer: Regularizer, X_train: lab.Tensor
) -> InternalModel:
    """Convert public-facing model objects into private implementations.

    Args:
        model: a model object from the public API.
        regularizer: a regularizer object from the public API.
        X_train: the :math:`n \\times d` training set.

    Returns:
        An internal model object with the same state as the public model.
    """
    assert isinstance(model, (LinearModel, ConvexReLU, ConvexGatedReLU))

    internal_model: InternalModel
    d, c = model.d + model.bias, model.c
    internal_reg = build_internal_regularizer(regularizer)

    if isinstance(model, LinearModel):
        return LinearRegression(d, c, regularizer=internal_reg)

    D, G = lab.all_to_tensor(
        compute_activation_patterns(
            lab.to_np(X_train), model.G, bias=model.bias
        )
    )

    if isinstance(model, ConvexReLU):
        internal_model = AL_MLP(
            d,
            D,
            G,
            "einsum",
            delta=1000,
            regularizer=internal_reg,
            c=c,
        )
    elif isinstance(model, ConvexGatedReLU):
        internal_model = ConvexMLP(
            d, D, G, "einsum", regularizer=internal_reg, c=c
        )
    else:
        raise ValueError(f"Model object {model} not supported.")

    return internal_model


def extract_bias(weights: lab.Tensor, bias: bool = False) -> List[np.ndarray]:
    """Extract optional bias columns from model weights.

    Args:
        weights: the weights or parameters of the model.
        bias: whether or not the weights contain a bias term.

    Returns
    """
    weights = lab.to_np(weights)

    if bias:
        return [weights[..., 0:-1], weights[..., -1]]
    else:
        return [weights]


def extract_gates_bias(
    G: lab.Tensor, bias: bool = False
) -> Tuple[np.ndarray, Optional[np.ndarray]]:

    G = lab.to_np(G)
    if bias:
        return (G[0:-1], G[-1])
    else:
        return (G, None)


def update_public_model(model: Model, internal_model: InternalModel) -> Model:
    """Update public-facing model object to match state of internal model.

    Args:
        model: the public-facing model.
        internal_model: the internal model object.

    Returns:
        The updated public-facing model.
    """
    # update neuron count
    model.p = internal_model.p

    if isinstance(model, ConvexGatedReLU):
        assert isinstance(internal_model, ConvexMLP)
        model.set_parameters(extract_bias(internal_model.weights, model.bias))
        model.G, model.G_bias = extract_gates_bias(
            internal_model.U, model.bias
        )
    elif isinstance(model, ConvexReLU):
        assert isinstance(internal_model, AL_MLP)
        model.set_parameters(
            extract_bias(internal_model.weights[0], model.bias)
            + extract_bias(internal_model.weights[1], model.bias)
        )

        model.G, model.G_bias = extract_gates_bias(
            internal_model.U, model.bias
        )
    elif isinstance(model, LinearModel):
        model.set_parameters(extract_bias(internal_model.weights, model.bias))

    return model


def build_public_model(
    internal_model: InternalModel,
    bias: bool = False,
) -> Model:
    """Construct a public-facing model from an internal model representation.

    Args:
        internal_model: the internal model.
        bias: whether or not the model contains a bias.

    Returns:
        A public-facing model with identical state.
    """
    model: Model
    if isinstance(internal_model, GatedReLUMLP):
        G, G_bias = extract_gates_bias(internal_model.U, bias)
        model = NonConvexGatedReLU(
            G, internal_model.c, bias=bias, G_bias=G_bias
        )
        w1, w2 = internal_model._split_weights(internal_model.weights)
        parameters = extract_bias(w1, bias) + extract_bias(w2, False)
        model.set_parameters(parameters)

    elif isinstance(internal_model, ReLUMLP):
        d = internal_model.d
        if bias:
            d = d - 1
        model = NonConvexReLU(d, internal_model.p, internal_model.c, bias=bias)
        w1, w2 = internal_model._split_weights(internal_model.weights)
        parameters = extract_bias(w1, bias) + extract_bias(w2, False)

        model.set_parameters(parameters)
    elif isinstance(internal_model, LinearRegression):
        model = LinearModel(internal_model.d, internal_model.c, bias=bias)
        model.parameters = extract_bias(internal_model.weights, bias)
    else:
        raise ValueError(f"Model {internal_model} not supported.")

    return model



"""L1 Regularizer."""

from typing import Optional

import lab

from scnn.private.loss_functions import l1_penalty
from scnn.private.models.regularizers.regularizer import Regularizer


class L1Regularizer(Regularizer):
    """The l1-regularizer, which takes the form.

    .. math:: r(w) = \\lambda \\|w\\||_2^2.

    This regularizer is non-smooth at :math:`w = 0`.

    Attributes:
        lam: the regularization strength. Must be non-negative.
    """

    def penalty(self, w: lab.Tensor, **kwargs) -> float:
        """Compute the penalty associated with the regularizer.

        Args:
            w: parameter at which to compute the penalty.
        """

        return l1_penalty(w, self.lam)

    def grad(
        self,
        w: lab.Tensor,
        base_grad: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the minimum-norm sub-gradient of the l1 regularizer.

        Args:
            w: parameter at which to compute the penalty gradient.
            base_grad: the smooth component of the gradient, coming from the objective function.

        Returns:
            The minimum-norm subgradient.
        """
        indicators = w != 0

        smooth_term = lab.sign(w) * self.lam * indicators

        non_smooth_term = (
            lab.sign(base_grad)
            * lab.smin(lab.abs(base_grad), self.lam)
            * lab.logical_not(indicators)
        )

        return smooth_term - non_smooth_term



"""Solution mappings for convex and non-convex formulations of neural network
training problems."""

from .mlps import (
    is_relu_compatible,
    is_grelu_compatible,
    construct_nc_manual,
    construct_nc_torch,
)

__all__ = [
    "is_relu_compatible",
    "is_grelu_compatible",
    "is_compatible",
    "get_nc_mlp",
    "get_nc_formulation",
    "is_compatible",
]

from typing import Union

import torch

from scnn.private.models.model import Model
from scnn.private.models.convex import (
    AL_MLP,
    ConvexMLP,
)
from scnn.private.models.linear import LinearRegression


def is_compatible(torch_model: torch.nn.Module) -> bool:
    """Check to see if there is a solution mapping mapping which is compatible
    with the architecture of the given model.

    :param torch_model: an instance of torch.nn.Module for which we want a convex program.
    :returns: true or false.
    """

    return is_relu_compatible(torch_model) or is_grelu_compatible(torch_model)


def get_nc_formulation(
    convex_model: Model,
    remove_sparse: bool = False,
) -> Union[torch.nn.Module, Model]:

    grelu = True
    if isinstance(convex_model, (AL_MLP)):
        grelu = False

    if isinstance(convex_model, ConvexMLP):
        return construct_nc_manual(convex_model, grelu, remove_sparse)
    elif isinstance(convex_model, LinearRegression):
        return convex_model
    else:
        raise ValueError(
            f"Model {convex_model} not recognized. Please add it to 'solution_mappings.mlps.py'"
        )



"""Interface for models.

TODO:
    - A better implementation for batching.
    - A better implementation for getting weights.
    - A better implementation for computing the objective scaling.
    - Rename `weights` to `parameters`.
    - Refactor `weights` to be a list of arrays, instead of a single array.
    """

from math import ceil
from typing import Optional, Tuple, Callable, Union, List, Dict

from scipy.sparse.linalg import (  # type: ignore
    LinearOperator,
    aslinearoperator,
)

import lab

from scnn.private.models.regularizers import Regularizer


class Model:

    """Interface for prediction models.

    Models classes wrap the functionality of a "forward pass", objective
    function, and "backward pass". Models must be able to

        1. Make predictions given data;

        2. Compute the objective associated with their predictions;

        3. Compute first-order derivatives of that objective.

    This abandons the compositionality of auto-differentiation software, but
    allows for optimized computation of the objective and gradient. A model
    has an optional regularizer, which computes some penalty on the model
    parameters.

    Attributes:
        d: the input dimension.
        c: the output dimension.
        weights: a tensor of parameters for the model.
        regularizer: an optional regularizer.
    """

    # public fields

    p: int = 0  # number of hidden units
    c: int = 1  # output dimension
    weights: lab.Tensor  # model parameters
    regularizer: Optional[Regularizer]  # penalty/regularizer

    # TODO: cleanup
    activation_history: Optional[lab.Tensor] = None
    weight_history: Optional[lab.Tensor] = None
    D: Optional[lab.Tensor] = None

    # private fields

    _train: bool = True

    def __init__(self, regularizer: Optional[Regularizer] = None):
        self.regularizer = regularizer

    def _scaling(self, y: lab.Tensor, scaling: Optional[float] = None):
        return scaling * self.c if scaling is not None else len(y)

    def batch_X(
        self, batch_size: Optional[int], X: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:

        if batch_size is None:
            return [{"X": X}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {"X": X[i * batch_size : (i + 1) * batch_size]}
            for i in range(n_batches)
        ]

    def batch_Xy(
        self, batch_size: Optional[int], X: lab.Tensor, y: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:

        if batch_size is None:
            return [{"X": X, "y": y}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "y": y[i * batch_size : (i + 1) * batch_size],
            }
            for i in range(n_batches)
        ]

    def eval(self):
        self._train = False

    def train(self):
        self._train = True

    def set_weights(self, weights: lab.Tensor):
        self.weights = weights

    def get_weights(self) -> List[lab.Tensor]:
        """Get model weights in an interpretable format.

        :returns: list of tensors.
        """
        return self.weights

    def _forward(self, X: lab.Tensor, w: lab.Tensor, **kwargs) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        :returns: the predictions f(x).
        """

        raise NotImplementedError(
            "A model must be associated with a objective function."
        )

    def _weights(self, w: Optional[lab.Tensor]) -> lab.Tensor:
        return self.weights if w is None else w

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> Union[float, lab.Tensor]:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the objective.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: objective L(f, (X,y)).
        """
        raise NotImplementedError(
            "A model must be associated with a objective function."
        )

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of the objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the gradient.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        raise NotImplementedError(
            "A model must be able to compute its gradient with respect to the objective."
        )

    def __call__(
        self,
        X: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: (optional) specific parameter at which to compute the forward pass.
            Defaults to 'None', in which case the current model state is used.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the predictions f(x).
        """
        return lab.concatenate(
            [
                self._forward(**batch, w=self._weights(w), **kwargs)
                for batch in self.batch_X(batch_size, X)
            ],
            axis=0,
        )

    def objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        ignore_regularizer: bool = False,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> Union[float, lab.Tensor]:
        """Compute objective associated with examples X and targets y.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: (optional) specific parameter at which to compute the objective.
            Defaults to 'None', in which case the current model state is used.
        :param ignore_regularizer: (optional) whether or not to ignore the regularizer and return
            *only* the un-regularized objective.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: objective L(f, (X,y)).
        """
        w = self._weights(w)

        obj = 0.0
        for batch in self.batch_Xy(batch_size, X, y):
            obj += self._objective(**batch, w=w, scaling=y.shape[0], **kwargs)

        if self.regularizer is not None and not ignore_regularizer:
            obj += self.regularizer.penalty(w)

        return obj

    def grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        ignore_regularizer: bool = False,
        return_model_grad: bool = False,
        batch_size: Optional[int] = None,
        step_size: Optional[float] = 1.0,
        **kwargs,
    ) -> Union[lab.Tensor, Tuple[lab.Tensor, lab.Tensor]]:
        """Compute the gradient of the objective with respect to the model
        parameters.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: (optional) specific parameter at which to compute the gradient.
            Defaults to 'None', in which case the current model state is used.
        :param ignore_regularizer: (optional) whether or not to ignore the regularizer and return
            *only* the gradient of the un-regularized objective.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :param step_size: (optional) step-size to use when computing the proximal gradient mapping.
            Defaults to 1.0.
        :returns: the gradient
        """
        w = self._weights(w)
        model_grad = lab.zeros_like(w)
        for batch in self.batch_Xy(batch_size, X, y):
            model_grad += self._grad(
                **batch, w=w, scaling=y.shape[0], **kwargs
            )

        grad = model_grad
        if self.regularizer is not None and not ignore_regularizer:
            grad += self.regularizer.grad(w, model_grad, step_size=step_size)

        if return_model_grad:
            return grad, model_grad

        return grad

    def get_closures(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        ignore_regularizer: bool = False,
        batch_size: Optional[int] = None,
    ) -> Tuple[Callable, Callable]:
        """Returns closures for computing the objective, gradient, and Hessian given (X, y).
            Warning: this closure will retain references to X, y and so can prevent garbage collection of
            these objects.
        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param ignore_regularizer: (optional) whether or not to ignore the regularizer and return
            closures for *only* the un-regularized objective.
        :returns: (objective_fn, grad_fn, hessian_fn).
        """

        def objective_fn(
            w: Optional[lab.Tensor] = None,
            ignore_regularizer=ignore_regularizer,
            batch_size=batch_size,
            **kwargs,
        ):
            return self.objective(
                X,
                y,
                w=w,
                ignore_regularizer=ignore_regularizer,
                batch_size=batch_size,
                **kwargs,
            )

        def grad_fn(
            w: Optional[lab.Tensor] = None,
            ignore_regularizer=ignore_regularizer,
            batch_size=batch_size,
            **kwargs,
        ):
            return self.grad(
                X,
                y,
                w=w,
                ignore_regularizer=ignore_regularizer,
                batch_size=batch_size,
                **kwargs,
            )

        return objective_fn, grad_fn

    def data_operator(self, X: lab.Tensor, **kargs) -> LinearOperator:
        """Construct a matrix operator to evaluate the data operator X v. where
        'v' is an input vector with shape (d,).

        :param X: (n,d) array containing the data examples.
        :returns: LinearOperator which computes the desired product.
        """
        return aslinearoperator(X)



"""Convert metrics from :module:`scnn.metrics` into internal
representations."""
from typing import Tuple, List, Dict, Any
from copy import deepcopy

import numpy as np

# Public Facing Objects

from scnn.metrics import Metrics


def build_metrics_tuple(
    metrics: Metrics,
) -> Tuple[List[str], List[str], List[str]]:
    """Convert Metrics instance into tuple of metrics for internal use.

    Args:
        metrics: object specifying which metrics should be collected during optimization.

    Returns:
        (train_metrics, test_metrics, additional_metrics) --- tuple of list of strings specifying which metrics should be collected.
    """
    train_metrics = []
    test_metrics = []
    additional_metrics = []

    for key, value in metrics.metrics_to_collect.items():
        if not value:
            continue

        if key == "objective":
            train_metrics.append("objective")
        elif key == "grad_norm":
            train_metrics.append("grad_norm")
        elif key == "time":
            continue
        elif key == "model_loss":
            train_metrics.append("base_objective")
        elif key == "constraint_gaps":
            train_metrics.append("constraint_gaps")
        elif key == "lagrangian_grad":
            train_metrics.append("lagrangian_grad")
        # use convex model for training (same as non-convex)
        elif key == "train_accuracy":
            train_metrics.append("accuracy")
        # use convex model for training (same as non-convex)
        elif key == "train_mse":
            train_metrics.append("squared_error")
        # use non-convex model for testing
        elif key == "test_accuracy":
            test_metrics.append("nc_accuracy")
        # use non-convex model for testing
        elif key == "test_mse":
            test_metrics.append("nc_squared_error")
        elif key == "active_neurons":
            additional_metrics.append("active_neurons")
        elif key == "neuron_sparsity":
            additional_metrics.append("group_sparsity")
        elif key == "active_features":
            additional_metrics.append("active_features")
        elif key == "feature_sparsity":
            additional_metrics.append("feature_sparsity")
        elif key == "active_weights":
            additional_metrics.append("active_weights")
        elif key == "weight_sparsity":
            additional_metrics.append("weight_sparsity")

    return train_metrics, test_metrics, additional_metrics


def update_public_metrics(
    metrics: Metrics, internal_metrics: Dict[str, Any]
) -> Metrics:
    """Update public-facing metrics object with optimization results.

    Args:
        metrics: a dictionary mapping metric names to data collected during optimization.
        internal_metrics: a public-facing metrics object.

    Returns:
        A new public-facing metrics object updated with metrics collected during optimization.
    """
    metrics = deepcopy(metrics)

    for key, value in internal_metrics.items():
        if key == "train_objective":
            metrics.objective = np.array(value)
        elif key == "train_grad_norm":
            metrics.grad_norm = np.array(value)
        elif key == "time":
            metrics.time = np.cumsum(np.array(value))
        elif key == "train_base_objective":
            metrics.model_loss = value
        elif key == "train_constraint_gaps":
            metrics.constraint_gaps = value
        elif key == "train_lagrangian_grad":
            metrics.lagrangian_grad = value
        elif key == "train_accuracy":
            metrics.train_accuracy = value
        # use convex model for training (same as non-convex)
        elif key == "train_squared_error":
            metrics.train_mse = value
        # use non-convex model for testing
        elif key == "test_nc_accuracy":
            metrics.test_accuracy = value
        # use non-convex model for testing
        elif key == "test_nc_squared_error":
            metrics.test_mse = value
        elif key == "active_neurons":
            metrics.active_neurons = value
        elif key == "group_sparsity":
            metrics.neuron_sparsity = value
        elif key == "active_features":
            metrics.active_features = value
        elif key == "feature_sparsity":
            metrics.feature_sparsity = value
        elif key == "active_weights":
            metrics.active_weights = value
        elif key == "weight_sparsity":
            metrics.weight_sparsity = value

    return metrics



"""Augmented lagrangian for the two-layer MLP with cone constraints."""
from math import ceil
from typing import Any, Dict, List, Optional, Tuple, Union

import lab

from scnn.private.loss_functions import squared_error
from scnn.private.models.convex import operators
from scnn.private.models.convex.convex_mlp import ConvexMLP
from scnn.private.models.regularizers import Regularizer


class AL_MLP(ConvexMLP):

    """Augmented Lagrangian for the two-layer ReLU MLP training problem."""

    def __init__(
        self,
        d: int,
        D: lab.Tensor,
        U: lab.Tensor,
        kernel: str = operators.EINSUM,
        delta: float = 100.0,
        regularizer: Optional[Regularizer] = None,
        c: int = 1,
    ) -> None:
        """
        :param d: the dimensionality of the dataset (ie. number of features).
        :param D: array of possible sign patterns.
        :param U: array of hyperplanes creating the sign patterns.
        :param c: number of classes for this problem
        :param kernel: the kernel to drive the matrix-vector operations.
        :param delta: parameter controlling the strength of the quadratic penalty
        in the augmented Lagrangian.
        """

        super().__init__(d, D, U, kernel=kernel, regularizer=regularizer, c=c)
        self.n = self.D.shape[0]
        self.c = c

        self.weights = lab.zeros((2, self.c, self.p, self.d))

        # estimates of the optimal dual variables
        self.e_multipliers = lab.tensor([0.0])  # NOT USED.
        self.i_multipliers = lab.zeros((2, self.c, self.p, self.n))

        self.delta = delta / self.n
        self.orthant = 2 * self.D - lab.ones_like(self.D)

    def _orthant(
        self, index_range: Optional[Tuple[int, int]] = None
    ) -> lab.Tensor:
        if index_range is None:
            return self.orthant

        return self.orthant[index_range[0] : index_range[1]]

    def _i_multipliers(
        self, index_range: Optional[Tuple[int, int]] = None
    ) -> lab.Tensor:
        if index_range is None:
            return self.i_multipliers

        return self.i_multipliers[:, :, :, index_range[0] : index_range[1]]

    def update_model_weights(self, w: lab.Tensor):
        self.weights = w

    def set_weights(self, weights: lab.Tensor):
        if weights.shape == (2, self.c, self.p, self.d):
            self.weights = weights
        elif weights.shape == (self.c, self.p, self.d):
            # "split" the weights and assign to the model
            self.weights = lab.stack([weights / 2, -weights / 2])
        else:
            raise ValueError(
                f"Weights with shape {weights.shape} cannot be set to IneqAugmentedConvexMLP with weight shape {self.weights.shape}."
            )

    def get_reduced_weights(self) -> lab.Tensor:
        return self.weights[0] - self.weights[1]

    def _forward(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        **kwargs,
    ) -> lab.Tensor:
        """Compute forward pass.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the forward pass.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :returns: predictions for X.
        """

        return super()._forward(X, w[0] - w[1], D)

    def lagrange_penalty_objective(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        **kwargs,
    ) -> float:
        """Compute the penalty term in the augmented Lagrangian.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the objective.
        :returns: the objective
        """
        w = w.reshape(2, self.c, self.p, self.d)

        gap = self.i_constraint_gap(X, w, index_range)
        scaled_multipliers = lab.safe_divide(
            self._i_multipliers(index_range), self.delta
        )

        penalty = (
            self.delta
            * lab.sum(lab.smax(gap + scaled_multipliers, 0) ** 2)
            / 2
        )

        return penalty

    def lagrange_penalty_grad(
        self,
        X: lab.Tensor,
        w: lab.Tensor,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        flatten=False,
        **kwargs,
    ) -> Union[Tuple[lab.Tensor, lab.Tensor], lab.Tensor]:
        """Compute the gradient of the penalty term in the augmented
        Lagrangian.

        :param X: (n,d) array containing the data examples.
        :param w: parameter at which to compute the gradient.
        :param flatten: whether or not to flatten the output into a single vector.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the gradient
        """
        w = w.reshape(2, self.c, self.p, self.d)
        gap = self.i_constraint_gap(X, w, index_range)

        scaled_multipliers = lab.safe_divide(
            self._i_multipliers(index_range),
            self.delta,
        )
        shifted_gap = lab.smax(gap + scaled_multipliers, 0)
        grad = -(
            self.delta
            * lab.einsum(
                "ij, lmji, ik->lmjk",
                self._orthant(index_range),
                shifted_gap,
                X,
            )
        )

        if flatten:
            return lab.ravel(grad)

        return grad

    def _objective(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        ignore_lagrange_penalty: bool = False,
        **kwargs,
    ) -> float:
        """Compute the l2 objective with respect to the model weights.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: specific parameter at which to compute the objective.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the objective
        """
        obj = squared_error(self._forward(X, w, D), y) / (
            2 * self._scaling(y, scaling)
        )
        if not ignore_lagrange_penalty:
            obj += self.lagrange_penalty_objective(X, w, index_range, scaling)

        return obj

    def _grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        ignore_lagrange_penalty: bool = False,
        flatten: bool = False,
        **kwargs,
    ) -> lab.Tensor:
        """Compute the gradient of augmented Lagrangian objective with respect
        to the model weights.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: parameter at which to compute the gradient.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :param flatten: whether or not to flatten the blocks of the gradient into a single vector.
        :returns: the gradient
        """
        w = w.reshape(2, self.c, self.p, self.d)
        combined_weights = w[0] - w[1]
        v_grad = super()._grad(X, y, combined_weights, D, scaling=scaling)

        grad = lab.stack([v_grad, -v_grad])
        if not ignore_lagrange_penalty:
            grad = grad + self.lagrange_penalty_grad(
                X, w, D=D, index_range=index_range, scaling=scaling
            )

        if flatten:
            return lab.ravel(grad)

        return grad

    def lagrangian(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: lab.Tensor,
        D: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        ignore_lagrange_penalty: bool = False,
        **kwargs,
    ) -> float:
        """Compute the Lagrangian function.

        :param X: (n,d) array containing the data examples.
        :param y: (n,d) array containing the data targets.
        :param w: specific parameter at which to compute the objective.
        :param D: (optional) specific activation matrix at which to compute the forward pass.
        Defaults to self.D or manual computation depending on the value of self._train.
        :param scaling: (optional) scaling parameter for the objective. Defaults to `n * c`.
        :returns: the objective
        """
        w = self._weights(w).reshape(2, self.c, self.p, self.d)

        # doesn't include regularization
        obj = squared_error(self._forward(X, w, D), y) / (
            2 * self._scaling(y, scaling)
        )

        # regularization
        if self.regularizer is not None:
            obj += self.regularizer.penalty(w)

        gap = self.i_constraint_gap(X, w, index_range)

        # penalty terms from Lagrangian
        penalty = lab.sum(gap * self._i_multipliers(index_range))

        return obj + penalty

    def lagrangian_grad(
        self,
        X: lab.Tensor,
        y: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        D: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
        scaling: Optional[float] = None,
        flatten: bool = False,
    ) -> lab.Tensor:
        w = self._weights(w).reshape(2, self.c, self.p, self.d)

        # Doesn't include regularization.
        obj_grad = self._grad(
            X,
            y,
            w,
            D,
            index_range,
            scaling,
            ignore_lagrange_penalty=True,
        )

        # penalty terms from Lagrangian
        penalty_grad = lab.einsum(
            "imjk, kj, kl -> imjl",
            self._i_multipliers(index_range),
            self._orthant(index_range),
            X,
        )

        grad = obj_grad - penalty_grad

        # regularization
        if self.regularizer is not None:
            grad += self.regularizer.grad(w, grad)

        if flatten:
            return lab.ravel(grad)

        return grad

    def i_constraint_gap(
        self,
        X: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
    ):
        """Compute violation of the linear constraints.

        .. math::

            X v_i - a_i

            X w_i - b_i,

        where a_i, b_i are the slack variables.
        :param X: (n,d) array containing the data examples.
        :param w: (optional) specific parameter at which to compute the gradient.
        Defaults to 'None', in which case the current model state is used.
        :returns: the constraint gaps.
        """
        weights = self._weights(w).reshape(2, self.c, self.p, self.d)
        gaps = -lab.einsum(
            "ij, lmkj, ik-> lmki", X, weights, self._orthant(index_range)
        )

        return gaps

    def constraint_gaps(
        self,
        X: lab.Tensor,
        w: Optional[lab.Tensor] = None,
        index_range: Optional[Tuple[int, int]] = None,
    ):
        """Compute violation of the constraints.

        :param X: (n,d) array containing the data examples.
        :param w: (optional) specific parameter at which to compute the gradient.
        Defaults to 'None', in which case the current model state is used.
        :returns: equality constraint gap, inequality constraint gap.
        """

        return lab.tensor([0.0]), lab.smax(
            self.i_constraint_gap(X, w, index_range), 0
        )

    def batch_Xy(
        self, batch_size: Optional[int], X: lab.Tensor, y: lab.Tensor
    ) -> List[Dict[str, lab.Tensor]]:
        D = self._signs(X)

        if batch_size is None:
            return [{"X": X, "y": y, "D": D}]

        n = X.shape[0]
        n_batches = ceil(n / batch_size)

        return [
            {
                "X": X[i * batch_size : (i + 1) * batch_size],
                "y": y[i * batch_size : (i + 1) * batch_size],
                "D": D[i * batch_size : (i + 1) * batch_size],
                "index_range": (i * batch_size, (i + 1) * batch_size),
            }
            for i in range(n_batches)
        ]

    def add_new_patterns(
        self, patterns: lab.Tensor, weights: lab.Tensor, remove_zero=False
    ) -> lab.Tensor:
        """Attempt to augment the current model with additional sign patterns.

        :param patterns: the tensor of sign patterns to add to the current set.
        :param weights: the tensor of weights which induced the new patterns.
        :param remove_zero: whether or not to remove the zero vector.
        :returns: None
        """

        if lab.size(patterns) > 0:
            # update neuron count.
            self.D = lab.concatenate([self.D, patterns], axis=1)
            self.U = lab.concatenate([self.U, weights.T], axis=1)

            # initialize new model components at 0.
            added_weights = lab.zeros((2, self.c, weights.shape[0], self.d))
            self.weights = lab.concatenate(
                [self.weights, added_weights], axis=2
            )

            # update dual variables
            added_multipliers = lab.zeros(
                (2, self.c, weights.shape[0], self.n)
            )
            self.i_multipliers = lab.concatenate(
                [self.i_multipliers, added_multipliers], axis=2
            )

            # filter out the zero column.
            if remove_zero:
                non_zero_cols = lab.logical_not(
                    lab.all(self.D == lab.zeros((self.D.shape[0], 1)), axis=0)
                )
                self.D = self.D[:, non_zero_cols]
                self.U = self.U[:, non_zero_cols]
                self.weights = self.weights[:, :, non_zero_cols]
                self.i_multipliers = self.i_multipliers[:, :, non_zero_cols]

            self.p = self.D.shape[1]

        self.orthant = 2 * self.D - lab.ones_like(self.D)

        return added_weights



"""Utilities for generating and processing datasets."""

from .synthetic import (
    gen_regression_data,
    gen_classification_data,
    gen_sparse_regression_problem,
)
from .transforms import unitize_columns, train_test_split, add_bias_col


__all__ = [
    "gen_classification_data",
    "gen_regression_data",
    "gen_sparse_regression_problem",
    "add_bias_col",
    "unitize_columns",
    "train_test_split",
]



"""Tools for training models with feature sparsity.


TODO:
    - Extract support from linear model.
    - Extract support by forward/backward.
"""
from typing import List, Tuple, Optional

import numpy as np
import lab

from scnn.regularizers import L1, FeatureGL1, Regularizer
from scnn.models import ConvexReLU, ConvexGatedReLU, LinearModel, Model
from scnn.solvers import RFISTA, Optimizer
from scnn.metrics import Metrics
from scnn.activations import sample_gate_vectors
from scnn.optimize import optimize_path, optimize_model

from scnn.private.utils.data.transforms import train_test_split


class SupportFinder:
    """Base class for support finders."""

    def __init__(self, valid_prop: float = 0.2):
        """ """
        self.valid_prop = valid_prop

    def __call__(
        self,
        model: Model,
        regularizer: Optional[Regularizer],
        solver: Optimizer,
        train_set: Tuple[np.ndarray, np.ndarray],
        device,
        dtype,
        seed,
    ) -> List[int]:
        raise NotImplementedError(
            "A SupportFinder must be callable to obtain \
                a support vector."
        )


class LinearSupportFinder(SupportFinder):
    """Find support by training a penalized linear model."""

    def __init__(self, lambda_path: List[float], valid_prop: float = 0.2):
        """ """
        self.lambda_path = lambda_path
        self.valid_prop = valid_prop

    def __call__(
        self,
        model: Model,
        regularizer: Optional[Regularizer],
        solver: Optimizer,
        train_set: Tuple[np.ndarray, np.ndarray],
        device,
        dtype,
        seed,
    ) -> List[int]:
        """ """
        path: List[Regularizer]
        X, y = train_set
        n, d = X.shape
        _, c = y.shape

        train, valid = train_test_split(
            train_set[0],
            train_set[1],
            valid_prop=self.valid_prop,
            split_seed=seed,
        )

        if c == 1:
            path = [L1(lam) for lam in self.lambda_path]
        else:
            path = [FeatureGL1(lam) for lam in self.lambda_path]

        model = LinearModel(d, c)
        solver = RFISTA(model)
        metrics = Metrics(train_mse=True, test_mse=True)

        path_models, path_metrics = optimize_path(
            model,
            solver,
            path,
            metrics,
            train[0],
            train[1],
            valid[0],
            valid[1],
            device=device,
            dtype=dtype,
            seed=seed,
        )

        # find best validation accuracy and extract support.
        best_mse = None
        best_index = None

        for i, m in enumerate(path_metrics):
            if best_mse is None or m.test_mse[-1] < best_mse:
                best_mse = m.test_mse[-1]
                best_index = i

        non_zeros = (
            lab.sum(lab.abs(path_models[best_index].parameters[0]), axis=0)
            != 0
        )

        support = lab.arange(d)[non_zeros].tolist()

        if len(support) == 0:
            # failure mode
            return lab.arange(d).tolist()
        else:
            return support


class ForwardBackward(SupportFinder):
    """Support finder based on the forward/backward method."""

    def __init__(self, forward: bool = True, valid_prop: float = 0.2):
        """ """

        self.valid_prop = valid_prop
        self.forward = forward

    def __call__(
        self,
        model: Model,
        regularizer: Optional[Regularizer],
        solver: Optimizer,
        train_set: Tuple[np.ndarray, np.ndarray],
        device,
        dtype,
        seed,
    ) -> List[int]:

        if not isinstance(model, (ConvexReLU, ConvexGatedReLU)):
            raise ValueError(
                "Forward-backward only supports convex reformulations."
            )

        p = model.p

        X, y = train_set
        n, d = X.shape
        _, c = y.shape

        train, valid = train_test_split(
            train_set[0],
            train_set[1],
            valid_prop=self.valid_prop,
            split_seed=seed,
        )

        remaining_features = np.arange(d).tolist()

        # find best validation accuracy and extract support.
        if self.forward:
            active_features = []
        else:
            active_features = remaining_features

        best_support = None
        best_crit = None

        for i in range(d):
            next_best_crit: float = None
            next_features: List[int] = None

            for feature in remaining_features:

                if self.forward:
                    features = active_features + [feature]
                else:
                    features = active_features.copy()
                    features.remove(feature)

                # Add or remove feature from dataset
                f_train = train[0][:, features], train[1]
                f_valid = valid[0][:, features], valid[1]

                # Update model (ie. resample gates).
                G = sample_gate_vectors(seed, len(features), p)
                f_model = model.__class__(G, model.c)

                # hard code metrics for now.
                metrics = Metrics(train_mse=True, test_mse=True)
                # train model
                f_model, f_metrics = optimize_model(
                    f_model,
                    solver,
                    metrics,
                    f_train[0],
                    f_train[1],
                    f_valid[0],
                    f_valid[1],
                    regularizer=regularizer,
                    return_convex=True,
                    device=device,
                    dtype=dtype,
                    seed=seed,
                )

                crit = f_metrics.test_mse[-1]

                # check criterion
                if next_best_crit is None or crit <= next_best_crit:
                    best_feature = feature
                    next_features = features
                    next_best_crit = crit

                if best_crit is None or crit <= best_crit:
                    best_support = features
                    best_crit = crit

            active_features = next_features
            remaining_features.remove(best_feature)

        return best_support



"""Process raw data into a format suitable for optimization.

TODO:
    - decide on the format for external data and provide this somewhere, possibly as part of a public `types` module.
"""

import lab

from scnn.private.utils.data.transforms import (
    unitize_columns,
    add_bias_col,
)


def normalized_into_input_space(
    weights: lab.Tensor, column_norms: lab.Tensor
) -> lab.Tensor:
    """Map the weights of a model in the column-normalized data space into the
    original data (ie. as input) space.

    Args:
        weights: a :math:`(\\ldots \\times d)` tensor of weights to map back into the input space.
        column_norms: a :math:`(d,)` vector of the column-norms of the original training data.

    Returns:
        A :math:`(\\ldots \\times d)` tensor of weights in the original
        input space.
    """
    return weights / column_norms


def input_into_normalized_space(model_weights, column_norms):
    """Map the weights of a model in the original data space into the column-
    normalized data space.

    Args:
        weights: a :math:`(\\ldots \\times d)` tensor of weights to map into the column-normalized space.
        column_norms: a :math:`(d)` vector of the column-norms of the original training data.

    Returns:
        A :math:`(\\ldots \\times d)` tensor of weights in the column-normalized space.
    """

    return model_weights * column_norms


def process_data(
    X_train,
    y_train,
    X_test,
    y_test,
    unitize_data: bool = True,
    bias: bool = False,
):
    """Process training and test data into a format suitable for optimization.

    The data can be input as a list of lists or in any format implementing a `to_list` method
    that can be called to obtain a list of lists.

    Args:
        X_train: :math:`(n \\times d)` matrix of training examples.
        y_train: :math:`(n)` or :math:`(n \\times c)` vector of training targets.
        X_test: :math:`(m \\times d)` matrix of test examples.
        y_test: :math:`(m)` or :math:`(n \\times c)` vector of test targets.
        unitize_data: whether or not to unitize the column norms of the
            training matrix.
        bias: whether or not to augment the data with a bias column.

    Returns:
        The input data as lab.Tensor instances with the correct datatype.
        Targets will be upcast into :math:`(n \\times 1)` matrices
        if the number of targets is 1 and a vector is provided.
    """
    # convert from input format into list of lists.

    X_train, y_train = [
        lab.tensor(to_list(v), dtype=lab.get_dtype())
        for v in [X_train, y_train]
    ]

    if X_test is None or y_test is None:
        assert X_test is None and y_test is None
        # spoof test data with training set
        X_test = X_train
        y_test = y_train
    else:
        X_test, y_test = [
            lab.tensor(to_list(v), dtype=lab.get_dtype())
            for v in [X_test, y_test]
        ]

    # add extra target dimension if necessary
    if len(y_train.shape) == 1:
        y_train = lab.expand_dims(y_train, axis=1)
        y_test = lab.expand_dims(y_test, axis=1)

    train_set, test_set, col_norms = (X_train, y_train), (X_test, y_test), None

    if bias:
        train_set, test_set = add_bias_col(train_set), add_bias_col(test_set)

    if unitize_data:
        train_set, test_set, col_norms = unitize_columns(train_set, test_set)

    return train_set, test_set, col_norms


def to_list(v):
    """Cast a vector or matrix or list of lists into a list of lists.

    Args:
        v: the vector, matrix, or list of cast. If `v` is not a list,
        it must implement a `tolist` method which can be called to
        convert `v` into a list.

    Returns:
        the argument `v` as a list of lists.
    """
    if isinstance(v, list):
        return v
    else:
        assert hasattr(v, "tolist")
        return v.tolist()



"""Regularizers for training neural networks by convex reformulation."""


class Regularizer:
    """Base class for all regularizers."""

    def __init__(self, lam: float):
        """Initialize the squared-error loss function."""
        self.lam = lam

    def __str__(self):
        return f"regularizer_{self.lam}"


class NeuronGL1(Regularizer):
    """A neuron-wise group-L1 regularizer.

    This regularizer produces neuron sparsity in the final model,
    meaning that some neurons will be completely inactive after training.
    The regularizer has the form,

    .. math:: R(U) = \\lambda \\sum_{i = 1}^p \\|U_i\\|_2,

    where :math:`\\lambda` is the regularization strength.

    Attributes:
        lam: the regularization strength.
    """

    def __str__(self):
        return f"neuron_gl1_{self.lam}"


class FeatureGL1(Regularizer):
    """A feature-wise group-L1 regularizer.

    This regularizer produces feature sparsity in the final model, meaning
    that some features will not be used after training.
    The regularizer has the form,

    .. math:: R(U) = \\lambda \\sum_{i = 1}^d \\|U_{\\cdot, i}\\|_2,

    where :math:`\\lambda` is the regularization strength.

    Attributes:
        lam: the regularization strength.
    """

    def __str__(self):
        return f"feature_gl1_{self.lam}"


class L2(Regularizer):
    """The standard squared-L2 norm regularizer, sometimes called weight-decay.

    The regularizer has the form,

    .. math:: R(U) = \\lambda \\sum_{i = 1}^p \\|U_{i}\\|^2_2,

    where :math:`\\lambda` is the regularization strength.

    Attributes:
        lam: the regularization strength.
    """

    def __str__(self):
        return f"l2_{self.lam}"


class L1(Regularizer):
    """The L1 norm regularizer.

    The regularizer has the form,

    .. math:: R(U) = \\lambda \\sum_{i = 1}^p \\|U_{i}\\|_1,

    where :math:`\\lambda` is the regularization strength.

    Attributes:
        lam: the regularization strength.
    """

    def __str__(self):
        return f"l1_{self.lam}"



"""Base class for optimization procedures."""

from logging import Logger
from typing import Dict, Any, List, Tuple, Callable, Optional
import timeit

import lab

from scnn.private.methods.optimizers import Optimizer
from scnn.private.models.model import Model
from scnn.private.metrics import (
    update_metrics,
    init_metrics,
    format_recent_metrics,
    merge_metric_lists,
)
from scnn.private.methods.termination_criteria import TerminationCriterion

# CONSTANTS

ITER_LOG_FREQ = 25
EPOCH_LOG_FREQ = 5

METRIC_FREQ = 1

# CLASSES


class OptimizationProcedure:

    """A single-call optimization procedure."""

    term_criterion: TerminationCriterion

    def __init__(
        self,
        optimizer: Optimizer,
        pre_process: Optional[Callable] = None,
        post_process: Optional[Callable] = None,
    ):
        """
        :param optimizer: an callable optimizer instance.
        """

        self.optimizer = optimizer
        self.pre_process = pre_process
        self.post_process = post_process

    def reset(self):
        """Reset the optimization procedure (including any attached optimizers)
        to their original state."""
        pass

    def _record_time(self, metrics_log, start_time):
        """Record time elapsed since start_time."""
        metrics_log["time"] = metrics_log.get("time", []) + [
            timeit.default_timer() - start_time
        ]

        return metrics_log, None

    def _get_start_time(self, start_time):
        return timeit.default_timer() if start_time is None else start_time

    def __call__(
        self,
        logger: Logger,
        model: Model,
        initializer: Callable[[Model], Model],
        train_data: Tuple[lab.Tensor, lab.Tensor],
        test_data: Tuple[lab.Tensor, lab.Tensor],
        metrics: Tuple[List[str], List[str], List[str]],
        final_metrics: Optional[Tuple[List[str], List[str], List[str]]] = None,
    ) -> Tuple[Dict[str, Any], Model, Dict[str, Any]]:
        """Optimize a model using an optimize method by directly calling an
        optimization procedure. Training and test metrics are collected before
        and after execution.

        :param logger: a logging.Logger object that can be used to log information during
            optimization.
        :param model: the model to optimize.
        :param initializer: a function that can be called to initialize the model.
        :param train_data: an (X, y) tuple representing the training set.
        :param test_data: an (X_test, y_test) tuple representing the test data.
        :param train_metrics: a list of strings indicating which metrics to evaluate on the
            training set.
        :param test_metrics: a list of strings indicating which metrics to evaluate on the
            test set.
        :param additional_metrics: a list of strings identifying the further metrics that should be computed.
        :returns: (exit_status, model, metrics) --- execution information, the optimized model, and metrics from optimization.
        """
        # setup
        X, y = train_data
        metrics_log: Dict[str, Any] = init_metrics(
            {}, merge_metric_lists(metrics, final_metrics)
        )

        start_time = self._get_start_time(None)
        model = initializer(model)

        # optional pre-processing.
        if self.pre_process is not None:
            model = self.pre_process(model, X, y)

        metrics_log, start_time = self._record_time(metrics_log, start_time)

        # pre-training metrics
        metrics_log = update_metrics(
            metrics_log, model, {}, train_data, test_data, metrics
        )

        logger.info(
            "Pre-Training Metrics: " + format_recent_metrics(metrics_log, metrics)
        )

        start_time = self._get_start_time(start_time)
        model, exit_status = self.optimizer(
            model,
            X,
            y,
        )

        # optional post-processing.
        if self.post_process is not None:
            model = self.post_process(model, X, y)

        # post-training metrics
        metrics_log, start_time = self._record_time(metrics_log, start_time)
        metrics_log = update_metrics(
            metrics_log,
            model,
            exit_status,
            train_data,
            test_data,
            merge_metric_lists(metrics, final_metrics),
        )

        logger.info(
            "Post-Training Metrics: " + format_recent_metrics(metrics_log, metrics)
        )

        return exit_status, model, metrics_log



"""Optimizers.

TODO:
    - Add optimal gradient method for L2-regularized and unregularized problems.
"""

# ===== module exports ===== #

from .optimizer import Optimizer
from .proximal_optimizer import ProximalOptimizer, ProximalLSOptimizer
from .meta_optimizer import MetaOptimizer
from .gd import GD, GDLS
from .pgd import PGD, PGDLS
from .fista import FISTA
from .augmented_lagrangian import AugmentedLagrangian

__all__ = [
    "Optimizer",
    "ProximalOptimizer",
    "ProximalLSOptimizer",
    "MetaOptimizer",
    "GD",
    "GDLS",
    "PGD",
    "PGDLS",
    "FISTA",
    "AugmentedLagrangian",
]



"""Implementation of augmented Lagrangian method for fitting two-layer MLPs
with orthant constraints."""
from typing import Dict, Any, Tuple, Optional

import lab

from scnn.private.models.model import Model
from scnn.private.methods.optimizers import Optimizer
from scnn.private.methods.termination_criteria import TerminationCriterion


class MetaOptimizer(Optimizer):

    """Meta optimization routine."""

    step_size: float = 0.0

    def step(
        self,
        model: Model,
        inner_term_criterion: TerminationCriterion,
        inner_optimizer: Optimizer,
        X: lab.Tensor,
        y: lab.Tensor,
        batch_index: Optional[int] = None,
    ) -> Tuple[Model, TerminationCriterion, Optimizer, Dict[str, Any]]:
        """Execute one step of a meta optimization routine.

        :param model: the model to update.
        :param inner_term_criterion: the termination criterion of the inner optimization
            routine to which this meta optimizer is being applied.
        :param inner_optimizer: the optimizer for the inner optimization
            routine to which this meta optimizer is being applied.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :returns: the updated model, inner termination criterion, inner optimizer,
            and exit status of any sub-procedures (ie. line-search)
        """

        raise NotImplementedError("A meta optimizer must implement 'step'!")

    def outer(self, model: Model) -> Model:
        """Put model in inner-loop mode.

        :param model: the model.
        :returns: updated model.
        """

        return model

    def inner(self, model: Model) -> Model:
        """Put model in outer-loop mode.

        :param model: the model.
        :returns: updated model.
        """

        return model



"""Loss functions and related utilities.

TODO:
    - Rewrite docstrings to be Google style and correct math format.
"""
from typing import Dict, Any, List, Tuple, Union, Optional

import lab

# loss functions


def squared_error(preds: lab.Tensor, y: lab.Tensor) -> float:
    """Compute squared loss,

        $(1/2)*||preds - y||_2^2$
    :param preds: predictions.
    :param y: targets.
    returns: squared-loss
    """
    residuals = y - preds
    return lab.sum((residuals) ** 2)


def logistic_loss(preds: lab.Tensor, y: lab.Tensor):
    """Compute the logistic loss,

        $log(1 + exp(-y*preds))$
    :param preds: predictions.
    :param y: targets. These should be in {-1, 1}
    :returns: the logistic loss
    """
    res = lab.logsumexp(lab.stack([lab.zeros_like(preds), -y * preds]), axis=0)
    return lab.sum(res)


def binned_accuracy(
    preds: lab.Tensor, y: lab.Tensor, n_classes: Optional[int] = 10
):
    """Compute the accuracy of classification method by binning the
    predictions.

    :param preds: the raw predictions from the model.
    :param y: the true targets/labels.
    :param n_classes: the number of target classes. It is assumed that the classes {0, 1, ..., n_classes}.
    :returns: accuracy score.
    """

    class_preds = lab.digitize(preds, lab.arange(n_classes) + 0.5)
    return lab.sum(class_preds == y) / y.shape[0]


def accuracy(preds: lab.Tensor, y: lab.Tensor):
    """Compute the accuracy of classification, assuming y is a one-hot encoded
    vector.

    :param preds: the raw predictions from the model.
    :param y: the one-hot encoded true targets/labels
    :returns: accuracy score
    """

    class_labels = lab.argmax(y, 1)
    class_preds = lab.argmax(preds, 1)
    return lab.sum(class_preds == class_labels) / y.shape[0]


# penalty functions


def l1_penalty(w: lab.Tensor, lam: float) -> float:
    """"""
    return lam * lab.sum(lab.abs(w))


def l2_penalty(w: lab.Tensor, lam: float) -> float:
    """"""

    return (lam / 2) * lab.sum(w ** 2)


def group_l1_penalty(w: lab.Tensor, lam: Union[lab.Tensor, float]) -> float:
    """Compute the penalty associated with the regularizer.

    :param w: the parameter at which to compute group l1 penalty. Note that 'w'
        must have shape (x, P), where 'P' is the number of groups.
    :param lam: the coefficient(s) controlling the strength of regularization.
        IF 'lam' is a numpy array, it must have shape (P,).
    :returns: penalty value
    """

    return lam * lab.sum(lab.sqrt(lab.sum(w ** 2, axis=-1)))


def l1_squared_penalty(w: lab.Tensor, lam: Union[lab.Tensor, float]) -> float:
    """Compute the penalty associated with the regularizer.

    :param w: the parameter at which to compute group l1 penalty. Note that 'w'
        must have shape (x, P), where 'P' is the number of groups.
    :param lam: the coefficient(s) controlling the strength of regularization.
        IF 'lam' is a numpy array, it must have shape (P,).
    :returns: penalty value
    """

    return (lam / 2) * lab.sum(lab.sum(lab.abs(w), axis=0) ** 2)


# "activation" functions


def relu(x: lab.Tensor) -> lab.Tensor:
    """Compute ReLU activation,

        $max(x, 0)$
    :param x: pre-activations
    """
    # what's the issue here?

    return lab.smax(x, 0)


def logistic_fn(x: lab.Tensor) -> lab.Tensor:
    """Compute logistic activation,

        $1 / (1 + exp(-x))$
    :param x: pre-activations
    """
    return 1.0 / (1.0 + lab.exp(-x))



"""Metrics that can be recorded while training models.
"""

import numpy as np


class Metrics(object):

    """Metrics collected while optimizing models.

    By default, only the objective, training time, and norm of the
    minimum-norm subgradient are collected.

    Attributes:
        objective: the optimization objective, including any constraint
            penalty terms.
        time: the cumulative wall-clock time.
        grad_norm: squared 2-norm of the minimum-norm subgradient of the
            optimization objective, including penalty terms.
        model_loss: the regularized loss of the model on the training set. This
            is the same as `objective` but does not include penalties.
        constraint_gaps: 2-norm of constraint violations.
            Ignored for unconstrained problems.
        lagrangian_grad: 2-norm of the (primal) subgradient of the
            Lagrangian function. Ignored for unconstrained problems.
        train_accuracy: accuracy on the training set.
        train_mse: mean squared error on the training set.
        test_accuracy": accuracy on the test set.
        test_mse: means squared error on the test set.
        total_neurons: total number of neurons (active and inactive) in the
            model.
        neuron_sparsity: proportion of neurons which are not used by the model
          (ie. all weights are exactly zero for those neurons).
        active_neurons: the number of neurons which are active (ie. the weights
            for those neurons are not exactly zero).
        total_features: total number of features in the dataset (both used and
            not used by the model).
        feature_sparsity: proportion of features which are not used by the
            model (ie. all weights are exactly zero for those features).
        active_features: the number of features currently used by the model.
        weight_sparsity: proportion of weights are which zero.
        metric_freq: the frequency (in iterations) at which metrics should be
            collected in iterations.
        metrics_to_collect: internal dictionary specifying which metrics
            should be collected.
    """

    test_metrics = False

    objective: np.ndarray
    grad_norm: np.ndarray
    time: np.ndarray
    model_loss: np.ndarray
    constraint_gaps: np.ndarray
    lagrangian_grad: np.ndarray
    train_accuracy: np.ndarray
    train_mse: np.ndarray
    test_accuracy: np.ndarray
    test_mse: np.ndarray
    total_neurons: np.ndarray
    neuron_sparsity: np.ndarray
    active_neurons: np.ndarray
    total_features: np.ndarray
    feature_sparsity: np.ndarray
    active_features: np.ndarray
    active_weights: np.ndarray
    weight_sparsity: np.ndarray

    def __init__(
        self,
        metric_freq: int = 25,
        objective: bool = True,
        grad_norm: bool = True,
        time: bool = True,
        model_loss: bool = False,
        constraint_gaps: bool = False,
        lagrangian_grad: bool = False,
        train_accuracy: bool = False,
        train_mse: bool = False,
        test_accuracy: bool = False,
        test_mse: bool = False,
        total_neurons: bool = False,
        neuron_sparsity: bool = False,
        active_neurons: bool = False,
        total_features: bool = False,
        feature_sparsity: bool = False,
        active_features: bool = False,
        active_weights: bool = False,
        weight_sparsity: bool = False,
    ):
        """
        Args:
            metric_freq: the frequency at which to log metrics.
            objective: the optimization objective, including any constraint
                penalty terms.
            time: the cumulative wall-clock time.
            grad_norm: squared 2-norm of the minimum-norm subgradient of the
                optimization objective, including penalty terms.
            model_loss: the regularized loss of the model on the training set.
                This is the same as `objective` but does not include penalties.
            constraint_gaps: 2-norm of constraint violations.
                Ignored for unconstrained problems.
            lagrangian_grad: 2-norm of the (primal) subgradient of the
                Lagrangian function. Ignored for unconstrained problems.
            train_accuracy: accuracy on the training set.
            train_mse: mean squared error on the training set.
            test_accuracy": accuracy on the test set.
            test_mse: means squared error on the test set.
            total_neurons: total number of neurons in the model.
            neuron_sparsity: proportion of neurons which are not used by the
                model (ie. all weights are exactly zero for those neurons).
            active_neurons: the number of neurons which are active (ie. the
                weights for those neurons are not exactly zero).
            total_features: total number of features in the dataset (both used
                and not used by the model).
            feature_sparsity: proportion of features which are not used by the
                model (ie. all weights are exactly zero for those features).
            active_features: number of features used by the model.
            active_features: number of features used by the model.
            weight_sparsity: proportion of weights are which zero.
        """

        self.metric_freq = metric_freq
        self.metrics_to_collect = {
            "objective": objective,
            "grad_norm": grad_norm,
            "time": time,
            "model_loss": model_loss,
            "constraint_gaps": constraint_gaps,
            "lagrangian_grad": lagrangian_grad,
            "train_accuracy": train_accuracy,
            "train_mse": train_mse,
            "test_accuracy": test_accuracy,
            "test_mse": test_mse,
            "total_neurons": total_neurons,
            "neuron_sparsity": neuron_sparsity,
            "active_neurons": active_neurons,
            "total_features": total_features,
            "feature_sparsity": feature_sparsity,
            "active_features": active_features,
            "active_weights": active_weights,
            "weight_sparsity": weight_sparsity,
        }

        if test_mse or test_accuracy:
            self.test_metrics = True

    def has_test_metrics(self) -> bool:
        """Returns `True` if any test-set metric is enabled, `False`
        otherwise."""
        return self.test_metrics



"""Proximal operators."""


from .proximal_ops import (
    ProximalOperator,
    Identity,
    L1,
    L2,
    GroupL1,
    FeatureGroupL1,
    Orthant,
    GroupL1Orthant,
)

__all__ = [
    "ProximalOperator",
    "Identity",
    "L1",
    "L2",
    "GroupL1",
    "FeatureGroupL1",
    "Orthant",
    "GroupL1Orthant",
]



"""Loss functions for training neural networks by convex reformulation.

Notes:
    - We only support the squared loss for the time being.
"""


class SquaredLoss:
    """The squared-error loss function.

    Given predictions :math:`f(X)` and targets :math:`y`, this loss function
    has the form,

    .. math:: L(f(X), y) = \\frac{1}{2} \\|f(X) - y\\|_2^2.
    """

    def __init__(self):
        """ """
        pass



"""Core subroutines for augmented Lagrangian methods and ADMM."""
from typing import Tuple

import numpy as np

import lab


def update_multipliers(
    e_multipliers: lab.Tensor,
    i_multipliers: lab.Tensor,
    e_gap: lab.Tensor,
    i_gap: lab.Tensor,
    delta: float,
) -> Tuple[lab.Tensor, lab.Tensor]:
    """
    :param e_multipliers: an estimate of the optimal dual parameters for the equality constraints.
    :param i_multipliers: an estimate of the optimal dual parameters for the inequality constraints.
    :param e_gap: the violation of the equality constraint.
    :param i_gap: the violation of the inequality constraint.
    :param delta: the strength of the quadratic penalty in the augmented Lagrangian.
    :returns: the updated dual parameters.
    """

    return e_multipliers + delta * e_gap, lab.smax(i_multipliers + delta * i_gap, 0)


def acc_update_multipliers(
    e_multipliers: lab.Tensor,
    i_multipliers: lab.Tensor,
    e_gap: lab.Tensor,
    i_gap: lab.Tensor,
    delta: float,
    e_v: lab.Tensor,
    i_v: lab.Tensor,
    t: float,
) -> Tuple[lab.Tensor, lab.Tensor, lab.Tensor, lab.Tensor, float]:
    """
    :param e_multipliers: an estimate of the optimal dual parameters for the equality constraints.
    :param i_multipliers: an estimate of the optimal dual parameters for the inequality constraints.
    :param e_gap: the violation of the equality constraint.
    :param i_gap: the violation of the inequality constraint.
    :param delta: the strength of the quadratic penalty in the augmented Lagrangian.
    :returns: the updated dual parameters.
    """

    t_plus = 1 + np.sqrt(1 + 4 * t ** 2) / 2
    beta = (t - 1) / t_plus

    e_plus, i_plus = e_v + delta * e_gap, lab.smax(i_v + delta * i_gap, 0)

    e_v_plus = e_plus + beta * (e_plus - e_multipliers)
    i_v_plus = i_plus + beta * (i_plus - i_multipliers)

    return e_plus, i_plus, e_v_plus, i_v_plus, t_plus



"""Line search conditions and tools."""
from typing import List

import numpy as np

import lab

# classes


class LSCondition:

    """Base class for line-search conditions."""

    def __call__(
        self,
        f0: float,
        f1: float,
        step: lab.Tensor,
        grad: lab.Tensor,
        step_size: float,
    ) -> bool:
        raise NotImplementedError(
            "Line-search conditions must implement '__call__'!"
        )


class FSS:

    """Base class for line-search conditions."""

    def __call__(
        self,
        f0: float,
        f1: float,
        step: lab.Tensor,
        grad: lab.Tensor,
        step_size: float,
    ) -> bool:
        return True


class QuadraticBound(LSCondition):
    def __call__(
        self,
        f0: float,
        f1: float,
        step: lab.Tensor,
        grad: lab.Tensor,
        step_size: float,
    ) -> bool:
        """Check the Armijo or sufficient progress condition, which is.

            $f(x_{k+1} ≤ f(x_k) + <grad, step> + eta/2 ||step||_2^2$.
        :param f0: previous objective value, f(x_k).
        :param f1: new objective value, f(x_{k+1}).
        :param step: the (descent) step to check against the Armijo condition,
            ie. the difference between the new iterate and the previous one,
                $step = x_{k+1} - x_k$.
        :param grad: the grad at which to check the Armijo condition. Must have the same shape as 'step'.
        :param step_size: the step-size used to generate the step.
        :returns: boolean indicating whether or not the Armijo condition holds.
        """

        step = lab.ravel(step)
        grad = lab.ravel(grad)

        return 2.0 * step_size * (f1 - f0 - lab.dot(step, grad)) < lab.dot(
            step, step
        )


class DiagQB(LSCondition):

    """ """

    def __init__(self, A: lab.Tensor):
        """
        :param rho: the relaxation parameter for the linearization. A sensible default is 1e-4.
        """
        self.A = A

    def __call__(
        self,
        f0: float,
        f1: float,
        step: lab.Tensor,
        grad: lab.Tensor,
        step_size: float,
    ) -> bool:
        """Check the Armijo or sufficient progress condition, which is.

            $f(x_{k+1} ≤ f(x_k) + <grad, step> + eta/2 ||step||_2^2$.
        :param f0: previous objective value, f(x_k).
        :param f1: new objective value, f(x_{k+1}).
        :param step: the (descent) step to check against the Armijo condition,
            ie. the difference between the new iterate and the previous one,
                $step = x_{k+1} - x_k$.
        :param grad: the grad at which to check the Armijo condition. Must have the same shape as 'step'.
        :param step_size: the step-size used to generate the step.
        :returns: boolean indicating whether or not the Armijo condition holds.
        """

        step = lab.ravel(step)
        grad = lab.ravel(grad)

        return 2.0 * step_size * (f1 - f0 - lab.dot(step, grad)) < lab.dot(
            step, self.A * step
        )


class Armijo(LSCondition):

    """The Armijo line-search condition."""

    def __init__(self, rho: float = 1e-4):
        """
        :param rho: the relaxation parameter for the linearization. A sensible default is 1e-4.
        """
        self.rho = rho

    def __call__(
        self,
        f0: float,
        f1: float,
        step: lab.Tensor,
        grad: lab.Tensor,
        step_size: float,
    ) -> bool:
        """Check the Armijo or sufficient progress condition, which is.

            $f(x_{k+1} ≤ f(x_k) + rho <grad, step>$.
        :param f0: previous objective value, f(x_k).
        :param f1: new objective value, f(x_{k+1}).
        :param step: the (descent) step to check against the Armijo condition,
            ie. the difference between the new iterate and the previous one,
                $step = x_{k+1} - x_k$.
        :param grad: the grad at which to check the Armijo condition. Must have the same shape as 'step'.
        :param step_size: the step-size used to generate the step.
        :returns: boolean indicating whether or not the Armijo condition holds.
        """

        return f1 <= f0 + self.rho * lab.sum(lab.multiply(step, grad))



"""Base class for iterative optimizers."""
from typing import Callable, Optional, Tuple, Dict, Any

import lab

from scnn.private.models.model import Model
from scnn.private.methods.line_search import (
    StepSizeUpdater,
    LSCondition,
    Backtracker,
)


class Optimizer:

    """The base class for optimization methods.

    Optimizer is a stateful wrapper around a 'step_fn' --- a function which can
    be called to compute one step of an iterative optimization method.
    """

    def __init__(
        self,
        step_fn: Callable,
        step_size: float,
        update_step_size: StepSizeUpdater,
    ):
        """
        :param step_fn: a function to call to execute one step of the iterative
            method. It should have the same signature as 'methods.gradient_descent.gd'.
        :param step_size: the constant step-size to use.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        self.step_fn = step_fn
        self.step_size = self.init_step_size = step_size
        self.old_step_size = step_size

        self.update_step_size = update_step_size

        self.f0: Optional[float] = None
        self.f1: Optional[float] = None

    def reset(self):
        """Reset the optimizer to its initial state."""

        self.step_size = self.init_step_size
        self.f0 = None
        self.f1 = None

    def step(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        f0: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
        batch_index: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> Tuple[Model, Optional[float], Dict[str, Any]]:
        """Execute one step of an iterative optimization method.

        :param model: the model to update.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param f0: (optional) current objective value. Can be provided to re-use computations.
        :param grad: (optional) current objective gradient. Can be provided to re-use computations.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: the updated model and exit status of any sub-procedures (ie. line-search)
        """
        # compute gradient and update model.
        if grad is None:
            grad = model.grad(X, y, batch_size=batch_size)

        old_weights = model.weights
        model.weights = self.step_fn(model.weights, grad, self.step_size)

        # update step-size
        self.step_size = self.update_step_size(
            self.step_size,
            self.old_step_size,
            model.weights - old_weights,
            self.f0,
            self.f1,
            grad,
        )

        return model, None, {"step_size": self.step_size}


class LSOptimizer(Optimizer):

    """Optimization method with line-search.

    Like optimizer, this is a stateful wrapper around a 'step_fn', which can be
    called to execute one step of the f.o. method, including the line-search.
    """

    def __init__(
        self,
        step_fn: Callable,
        init_step_size: float,
        ls_cond: LSCondition,
        backtrack_fn: Backtracker,
        update_step_size: StepSizeUpdater,
    ):
        """
        :param step_fn: a function to call to execute one step of the iterative
            method. It should have the same signature as 'methods.gradient_descent.gd_ls'.
        :param init_step_size: first step-size to try when running the line-search.
        :param ls_cond: the line-search condition to check.
        :param backtrack: a rule for calculating the next step-size to try.
        :param update_step_size: a rule for updating the step-size after each line-search.
        """
        super().__init__(step_fn, init_step_size, update_step_size)

        self.ls_cond = ls_cond
        self.backtrack_fn = backtrack_fn

    def reset(self):
        """Reset the optimizer to its initial state."""
        super().reset()
        self.old_step_size = None

    def step(
        self,
        model: Model,
        X: lab.Tensor,
        y: lab.Tensor,
        f0: Optional[float] = None,
        grad: Optional[lab.Tensor] = None,
        batch_index: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> Tuple[Model, Optional[float], Dict[str, Any]]:
        """Execute one step of an iterative optimization method.

        :param model: the model to update.
        :param X: the (n,d) data matrix to use in the update.
        :param y: the (n,) vector of targets.
        :param f0: (optional) current objective value. Can be provided to re-use computations.
        :param grad: (optional) current objective gradient. Can be provided to re-use computations.
        :param batch_index: NOT USED. The index of the current mini-batch.
        :param batch_size: the batch size to use when computing the objective.
            Defaults to `None' which indicates full-batch.
        :returns: (model, status_dict) --- the updated model and a dictionary describing the optimizer's exit state.
        """

        obj_fn, grad_fn = model.get_closures(X, y, batch_size=batch_size)

        if grad is None:
            grad = grad_fn(model.weights)

        if self.f1 is None:
            self.f0 = obj_fn(model.weights)
        else:
            self.f0 = self.f1

        old_weights = model.weights
        model.weights, self.f1, self.step_size, exit_state = self.step_fn(
            model.weights,
            self.f0,
            grad,
            grad,
            obj_fn,
            grad_fn,
            self.step_size,
            self.ls_cond,
            self.backtrack_fn,
            prox=None,
        )

        # update step-size
        self.step_size = self.update_step_size(
            self.step_size,
            self.old_step_size,
            model.weights - old_weights,
            self.f0,
            self.f1,
            grad,
        )
        self.old_step_size = self.step_size

        return model, self.f1, exit_state
